import json
import re
import ast
from copy import deepcopy


def add_column_based_on_null(event, field, new_field, new_value_if_null, new_value_if_not_null):
    """ Checks and adds a value to a new field based on NULL

    :param event: A dictionary
    :param field: The name of the field to check
    :param new_field: The name of the new field
    :param new_value_if_null: The value for the new_field if field IS NULL
    :param new_value_if_not_null: The value for the new_field if field IS NOT NULL
    :return: An altered dictionary

    Examples:

    .. code-block:: python

        # Example #1
        event = {'do i have digits?': '1234'}
        event = add_column_based_on_null(event, field='do i have digits?', new_field='digits', new_value_if_null='N', new_value_if_not_null='Y')
        event = {'digits': 'Y',
                'do i have digits?': '1234'}

        # Example #2
        event = {'do i have digits?': None}
        event = add_column_based_on_null(event, field='do i have digits?', new_field='digits', new_value_if_null='N', new_value_if_not_null='Y')
        event = {'digits': 'N',
                'do i have digits?': None}

    """
    if field in event:
        if event[field] is None:
            event[new_field] = new_value_if_null
        else:
            event[new_field] = new_value_if_not_null

    return event


def add_columns_with_default(event, field_and_default_dict):
    """ Adds a column with the default value to every event where it's not already present

    :param event: A dictionary
    :param field_and_default_dict: A dictionary with keys and default values {field_1: default_for_field_1, field_2: default_for_field_2}
    :return: An altered dictionary

    Examples:

    .. code-block:: python

        # Example #1
        event = {'an existing column': 'some value'}
        event = add_columns_with_default(event, field_and_default_dict={"new column 1": 'my default of 1', "new column two": 2})
        event = {'an existing column': 'some value',
                'new column 1': 'my default of 1',
                'new column two': 2}

        # Example #2
        event = {'an existing column': 'some value'}
        event = add_columns_with_default(event, field_and_default_dict={"mark_for_delete": False})
        event = {'an existing column': 'some value',
                'mark_for_delete': False}

    """

    for k, v in field_and_default_dict.items():
        if k not in event:
            event[k] = v
    return event


def add_composite_key(event, field_list, key_name):
    """ Creates a composite key to be used for a unique ID.

    :param event: A dictionary
    :param field_list: A list of fields to combine to make the key
    :param key_name: The name of the new key field
    :return: An altered dictionary

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': 1234,
                'a_second_field': '2019-01-01',
                'a_third_field': 'abc'}
        event = add_composite_key(event, field_list=['a_field', 'a_second_field', 'a_third_field'], key_name='my_derived_key')
        event = {'a_field': 1234,
                'a_second_field': '2019-01-01',
                'a_third_field': 'abc',
                'my_derived_key':
                '1234-2019-01-01-abc'}

    """

    key_list = []
    for field in field_list:
        key_list.append(event[field])

    if None in key_list:
        raise Exception('Missing values in composite key')
    else:
        key_list = [str(x) for x in key_list]
        event[key_name] = '-'.join(key_list)
    return event


def add_duplicate_fields(event, field_name, suffix_or_suffix_list, keep_original=False):
    """ Add duplicate values of a field with a suffix

    :param event: A dictionary
    :param field_name: The name of the field to duplicate
    :param suffix_or_suffix_list: A single or list of suffixes to add to the field_name in the duplicates
    :param keep_original: True to keep the original field also
    :return: An altered dictionary

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': 'a_value'}
        event = add_duplicate_fields(event, field_name='a_field', suffix_or_suffix_list='my_suffix', keep_original=False)
        event = {'a_field_my_suffix': 'a_value'}

        # Example #2
        event = {'a_field': 'a_value'}
        event = add_duplicate_fields(event, field_name='a_field', suffix_or_suffix_list='my_suffix', keep_original=True)
        event = {'a_field': 'a_value',
                'a_field_my_suffix': 'a_value'}

        # Example #3
        event = {'a_field': 'a_value'}
        event = add_duplicate_fields(event, field_name='a_field', suffix_or_suffix_list=['my_suffix','my_second_suffix'], keep_original=False)
        event = {'a_field_my_second_suffix': 'a_value',
                'a_field_my_suffix': 'a_value'}

    """

    if type(suffix_or_suffix_list) is not list:
        suffix_or_suffix_list = [suffix_or_suffix_list]

    if field_name in event:
        for suffix in suffix_or_suffix_list:
            event[field_name + '_' + suffix] = event[field_name]

    if not keep_original:
        del event[field_name]

    return event


def add_suffix(event, fields, suffix, separator='_'):
    """ Adds a suffix to a field or list of fields

    :param event: A dict with the entire event
    :param field_or_field_list: A single field or list of fields for which to add a suffix
    :param suffix: The suffix to add to the fields
    :param separator: The character to place between the name and the suffix
    :return: An altered event with the altered field names

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': 'a_value'}
        event = add_suffix(event, fields='a_field', suffix='an_ending')
        event = {'a_field_an_ending': 'a_value'}

        # Example #2
        event = {'a_field': 'a_value'}
        event = add_suffix(event, fields='a_field', suffix='an_ending', separator='---')
        event = {'a_field---an_ending': 'a_value'}

        # Example #3
        event = {'a_field': 'a_value',
                'another_field': 'another_value'}
        event = add_suffix(event, fields=['a_field', 'another_field'], suffix='an_ending')
        event = {'a_field_an_ending': 'a_value',
                'another_field_an_ending': 'another_value'}

    """

    if type(fields) is not list:
        fields = [fields]

    for k, v in event.items():
        if k in fields:
            event[k + separator + suffix] = event[k]
            del event[k]

    return event


def convert_all_event_fields_to_snake_case(event):
    """ Converts all keys in an event to snake case.
    If a key is Partially_SnakeCase we'll get 2 underscores where there is currently one like partially__snake_case

    :param event: An Alooma event
    :return: A transformed event with all the keys in snake_case

    Examples:

    .. code-block:: python

        # Example #1
        event = {'_metadata': {},
                'TitleCase': 'to_snake_case',
                'camelCase': 'to_snake_case',
                'snake_case': 'to_snake_case'}
        event = convert_all_event_fields_to_snake_case(event)
        event = {'_metadata': {},
                'camel_case': 'to_snake_case',
                'snake_case': 'to_snake_case',
                'title_case': 'to_snake_case'}


    """

    new_event = {}

    for k, v in event.items():
        new_event[convert_string_to_snake_case(k).replace(' ', '_')] = v

    new_event['_metadata'] = deepcopy(event['_metadata'])
    return new_event


def convert_dictionary_fields_to_string(event, field_or_field_list):
    """ Dumps a list of fields to a string to keep Alooma from auto-parsing

    :param event: A dict with the entire event
    :param field_or_field_list: A single field or list of fields to json.dumps to keep Alooma from doing infinte de-nesting
    :return: A new event

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_dict': {'field_one': 1,
                'field_two': 2}}
        event = convert_dictionary_fields_to_string(event, field_or_field_list='a_dict')
        event = {'a_dict': '{"field_one": 1, "field_two": 2}'}

        # Example #2
        event = {'a_dict': {'field_one': 1,
                            'field_two': 2},
                'a_second_dict': {'field_one': 1,
                                'field_two': 2}}
        event = convert_dictionary_fields_to_string(event, field_or_field_list=['a_dict', 'a_second_dict'])
        event = {'a_dict': '{"field_one": 1, "field_two": 2}',
                'a_second_dict': '{"field_one": 1, "field_two": 2}'}

    """

    if type(field_or_field_list) is not list:
        field_or_field_list = [field_or_field_list]

    for field in field_or_field_list:
        if field in event and event[field] is not None:
            try:
                event[field] = json.dumps(event[field])
            except:
                pass

    return event


def convert_null_to_zero(event, field_or_field_list):
    """ Converts the value in a field or field list from None to 0

    :param event: a dict with the event
    :param field_or_field_list: A single field or list of fields to convert to 0 if null
    :return: the updated event

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': None}
        event = convert_null_to_zero(event, field_or_field_list='a_field')
        event = {'a_field': 0}

        # Example #2
        event = {'a_field': None,
                'another_field': None}
        event = convert_null_to_zero(event, field_list=['a_field', 'another_field'])
        event = {'a_field': 0,
                'another_field': 0}

    """

    if type(field_or_field_list) is str:
        field_or_field_list = [field_or_field_list]

    for field in field_or_field_list:
        if field in event and event[field] is None:
            event[field] = 0

    return event


def convert_spaces_and_special_characters_to_underscore(name):
    """ Converts spaces and special characters to underscore so 'Thi$ i# jun&' becomes 'thi__i__jun_'

    :param name: A string
    :return: An altered string

    Example use case:
    - A string might have special characters at the end when they are really the same field such as  My_Field$ and My_Field#
    - We use this to covert the names to "my_field" to combine the values so the events will be easily grouped together

    Examples:

    .. code-block:: python

        # Example #1
        input_string = '$Scr "get-rid^-of-the@" special #characters%&space'
        output_string = convert_spaces_and_special_characters_to_underscore(input_string)
        output_string = '_scr__get_rid__of_the___special__characters__space'

    """

    clean_name = re.sub(r'[\W_]', '_', name)
    return clean_name.lower()


def convert_string_to_snake_case(name):
    """ Converts a string to Snake Case

    :param name: A string
    :return: A string in snake case

    Example use case:
    - Events from might have custom properties in camelCase like userId, and userEmail
    - Use this to rename keys to user_id and user_email for better ease of reading in the database

    Examples:

    .. code-block:: python

        # Example #1
        input_string = 'MakeThisSnakeCase'
        output_string = convert_string_to_snake_case(input_string)
        output_string = 'make_this_snake_case'

        # Example #2
        input_string = 'Make This Snake Case'
        output_string = convert_string_to_snake_case(input_string)
        output_string = 'make_this_snake_case'

        # Example #3
        input_string = 'keep_this_snake_case'
        output_string = convert_string_to_snake_case(input_string)
        output_string = 'keep_this_snake_case'

    """

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    sl = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
    return sl.replace(' ', '')


def convert_values_to_none(event, field_or_field_list, field_values=None):
    """ Changes a field to None. If a field value is specified then only that value will be changed to None

    :param event: A dictionary
    :param field_or_field_list: A single field or list of fields  to convert to None
    :param field_values: The value to convert to None. If specified only these values are converted to None
    :return: An altered dictionary

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': 'a_value'}
        event = convert_values_to_none(event, field_or_field_list='a_field')
        event = {'a_field': None}

        # Example #2
        event = {'a_field': 'a_value',
                'another_field': 'another_value'
        event = convert_values_to_none(event, field_or_field_list=['a_field', 'another_field'])
        event = {'a_field': None,
                'another_field': None}

        # Example #3
        event = {'a_field': 'a_value',
                'another_field': 'another_value'
        event = convert_values_to_none(event, fields=['a_field', 'another_field'], field_values='a_value')
        event = {'a_field': None,
                'another_field': 'another_value'}

        # Example #4
        event = {'a_field': 'a_value',
                'another_field': 'another_value'
        event = convert_values_to_none(event, fields=['a_field', 'another_field'], field_values=['a_value', 'another_value'])
        event = {'a_field': None,
                'another_field': None}

    """

    if type(field_or_field_list) is not list:
        field_or_field_list = [field_or_field_list]

    if field_values is not None:
        if type(field_values) is not list:
            field_values = [field_values]

    for field in field_or_field_list:
        if field in event and field_values is None:
            event[field] = None
        elif field in event and event[field] in field_values:
            event[field] = None

    return event


def convert_empty_value_to_none(event, key_name):
    """ Changes an empty string of "" or " ", and empty list of [] or an empty dictionary of {} to None so it will be NULL in the database

    :param event: A dictionary
    :param key_name: The key for which to check for empty strings
    :return: An altered dictionary

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': ' '}
        event = convert_empty_value_to_none(event, key_name='a_field')
        event = {'a_field': None}

        # Example #2
        event = {'a_field': '{}'}
        event = convert_empty_value_to_none(event, key_name='a_field')
        event = {'a_field': None}

        # Example #3
        event = {'a_field': {}}
        event = convert_empty_value_to_none(event, key_name='a_field')
        event = {'a_field': None}

    """

    if key_name in event:
        if type(event[key_name]) == str and (event[key_name] == '' or event[key_name].strip() == '' or event[key_name] == '{}' or event[key_name] == '[]'):
            event[key_name] = None

        # Converts an empty list or dictionary to None
        if not event[key_name]:
            event[key_name] = None

    return event


def convert_event_type_case(event, case_force_upper=False):
    """ Forces upper of lower case for event types at the end of the code engine.  For Snowfalke force UPPER and for Redshift force lower.

    :param event: A dict with the entire event
    :param case_force_upper: True to for upper case.
    :return: An event with the case altered in event_type

    Examples:

    .. code-block:: python

        # Example #1
        event = {'_metadata': {'event_type': 'My_SCHEMA.my_table'}}
        event = convert_event_type_case(event)
        event = {'_metadata': {'event_type': 'my_schema.my_table'}}

        # Example #2
        event = {'_metadata': {'event_type': 'My_SCHEMA.my_table'}}
        event = convert_event_type_case(event, case_force_upper=True)
        event = {'_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'}}

    """

    if case_force_upper:
        if type(event) == list:
            for each_event in event:
                if 'event_type' in each_event['_metadata']:
                    each_event['_metadata']['event_type'] = each_event['_metadata']['event_type'].upper()
        else:
            if 'event_type' in event['_metadata']:
                event['_metadata']['event_type'] = event['_metadata']['event_type'].upper()
    else:
        if type(event) == list:
            for each_event in event:
                if 'event_type' in each_event['_metadata']:
                    each_event['_metadata']['event_type'] = each_event['_metadata']['event_type'].lower()
        else:
            if 'event_type' in event['_metadata']:
                event['_metadata']['event_type'] = event['_metadata']['event_type'].lower()

    return event


def flatten_json(event, field_or_field_list, levels, keep_original=False, dump_to_string=False):
    """ Flattens a list of fields from a dictionary n levels

    :param event: the event that you want to pass through the function (formatted as a dictionary)
    :param field_or_field_list: A field or list of the fields that you want to flatten N levels deep
    :param levels: The number of levels that you want to parse the fields
    :param keep_original: True if you want to keep the original field in the event, false if you want to delete it
    :return:  The transformed event

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_dict': {'field_one': 1,
                            'field_two': 2}}
        event = flatten_json(event, field_or_field_list='a_dict', levels=1)
        event = {'a_dict_field_one': 1,
                'a_dict_field_two': 2}

        # Example #2
        event = {'a_dict': {'field_one': 1,
                            'field_two': 2},
                'a_second_dict': {'field_one': {'one more': 1.1,
                                                'one more again': 1.2},
                                'field_two': 2}}
        event = flatten_json(event, field_or_field_list=['a_dict','a_second_dict'], levels=1)
        event = {'a_dict_field_one': 1,
                'a_dict_field_two': 2,
                'a_second_dict_field_one': {'one more': 1.1, 'one more again': 1.2},
                'a_second_dict_field_two': 2}

        # Example #3
        event = {'a_dict': {'field_one': 1,
                            'field_two': 2},
                'a_second_dict': {'field_one': {'one more': 1.1,
                                                'one more again': 1.2},
                                'field_two': 2}}
        event = flatten_json(event, field_or_field_list=['a_dict','a_second_dict'], levels=2)
        event = {'a_dict_field_one': 1,
                'a_dict_field_two': 2,
                'a_second_dict_field_one_one more': 1.1,
                'a_second_dict_field_one_one more again': 1.2,
                'a_second_dict_field_two': 2}

    """

    # Turns a single field into a list
    if type(field_or_field_list) is str:
        field_or_field_list = [field_or_field_list]

    for field in field_or_field_list:
        field = field.lower()
        if field in event.keys():

            # make a copy of the original event to be adding back to event if specified
            original_copy = event[field]

            counter = 0
            pre_fix_list = []
            pre_fix_list.append(field)

            while counter < levels:

                next_level_pre_fix_list = []
                for pre_fix in pre_fix_list:
                    pre_fix_underscore = pre_fix.lower() + '_'

                    event = flatten_json_1_level(event, field_name=pre_fix, field_name_underscore=pre_fix_underscore, dump_to_string=dump_to_string)
                    next_level_pre_fix_list.append(pre_fix_underscore)

                for n in next_level_pre_fix_list:
                    for k in event.keys():

                        if str(k).startswith(n):
                            pre_fix_list.append(k)

                    pre_fix_list = [p for p in pre_fix_list if p != n[:-1]]

                counter += 1

            if keep_original:
                if type(original_copy) is dict and original_copy != {} and dump_to_string:
                    event[field] = json.dumps(original_copy)
                elif dump_to_string:
                    event[field] = original_copy
                else:
                    try:
                        event[field] = json.loads(original_copy)
                    except:
                        try:
                            event[field] = ast.literal_eval(original_copy)
                        except:
                            event[field] = original_copy

    return event


def flatten_json_1_level(event, field_name, field_name_underscore, dump_to_string):
    """ Flattens a JSON field 1 level. This function is used in flatten JSON

    :param event: A dictionary
    :param field_name: The field name to flatten
    :param field_name_underscore: The field name with an underscore appended
    :param dump_to_string: If true any remaining dictionaries will be converted to a string with json.dumps
    :return: An event with the field flattened

    Examples:

    .. code-block:: python

        # Example #1
        event = {'my_field': "{\"a\": None, \"b\"}"}
        event = flatten_json_1_level(event=input_event, field_name='my_field', field_name_underscore='my_field_', dump_to_string=True)
        output_event = {'my_field_a': None,
                        'my_field_b': 2}

    """

    # Load strings to JSON when possible, otherwise return the event
    if type(event[field_name]) is not dict:
        try:
            event[field_name] = json.loads(event[field_name])
        except:
            try:
                event[field_name] = ast.literal_eval(event[field_name])
            except:
                return event

    # iterate through the dictionary and flatten a single level
    try:
        for k, v in event[field_name].items():
            if type(v) is dict and dump_to_string:
                event[field_name_underscore + k.lower()] = json.dumps(v)
            else:
                event[field_name_underscore + k.lower()] = v

        del event[field_name]

    except:
        return event

    return event


def map_key_in_dictionary_to_value(event, mapping_dict, existing_column, new_column, allow_nulls):
    """ Adds a column mapping using a dictionary

    :param event: A dictionary
    :param mapping_dict: A mapping dict such as {1: 'product A', 2: 'product B'}
    :param existing_column:  The column that matches the keys in the mapping dict
    :param new_column: The name of the column to put the values from the mapping dict
    :param allow_nulls: True if the function should let a NULL value in the existing_column pass through. False to throw an error when the existing column has NULL.
    :return: An event with the new_column k, v added

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': 1,
                '_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'}}
        event = map_key_in_dictionary_to_value(event, mapping_dict={1: 'one', 2: 'two'}, existing_column='a_field', new_column='a_mapped_field', allow_nulls=False)
        event = {'_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'},
                'a_field': 1,
                'a_mapped_field': 'one'}

        # Example #2
        event = {'a_field': 3,
                '_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'}}
        event = map_key_in_dictionary_to_value(event, mapping_dict={1: 'one', 2: 'two'}, existing_column='a_field', new_column='a_mapped_field', allow_nulls=False)
        Exception: Missing enum transform MY_SCHEMA.MY_TABLE a_field

        # Example #3
        event = {'a_field': None,
                '_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'}}
        event = map_key_in_dictionary_to_value(event, mapping_dict={1: 'one', 2: 'two'}, existing_column='a_field', new_column='a_mapped_field', allow_nulls=True)
        event = {'_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'},
                'a_field': None,
                'a_mapped_field': None}

    """

    func_name = 'enum transform ' + event['_metadata']['event_type'] + ' ' + existing_column

    existing_column_value = None

    if existing_column in event:
        existing_column_value = event[existing_column]

    if allow_nulls and (existing_column_value == None or existing_column_value in mapping_dict.keys()):
        event[new_column] = mapping_dict.get(existing_column_value, None)

    elif not allow_nulls and event[existing_column] in mapping_dict.keys():
        event[new_column] = mapping_dict.get(existing_column_value, None)

    else:
        raise Exception('Missing %s' % func_name)

    return event


def map_value_in_list_to_dictionary_key(event, mapping_dict_with_lists, existing_column, new_column, allow_nulls, passthrough):
    """  Maps a value from a list back to the key. Useful to map values to categories.

    :param event: A dictionary
    :param mapping_dict_with_lists:  A mapping dict with lists in the values such as {"baked good": ["cake", "croissant"]}
    :param existing_column: The column with the existing data
    :param new_column: The name of the new column for the added data
    :param allow_nulls: True if the existing column can have NULL. If set to False NULLs will throw an error
    :param passthrough: True if we should pass through a value of the existing column when there is no mapping value in the list
    :return: An altered event

    Examples:

    .. code-block:: python

        # Example #1
        event = {'_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'},
                'a_field': 1}
        event = map_value_in_list_to_dictionary_key(event, mapping_dict_with_lists={'1-3': [1, 2, 3], '4-6': [4, 5, 6]}, existing_column='a_field', new_column='a_mapped_field', allow_nulls=False, passthrough=False)
        event = {'_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'},
                'a_field': 1,
                'a_mapped_field': '1-3'}

        # Example #2
        event = {'_metadata': {'event_type': 'MY_SCHEMA.MY_TABLE'},
                'a_field': 7}
        event = map_value_in_list_to_dictionary_key(event, mapping_dict_with_lists={'1-3': [1, 2, 3], '4-6': [4, 5, 6]}, existing_column='a_field', new_column='a_mapped_field', allow_nulls=False, passthrough=False)
        Exception: Missing map_list transform MY_SCHEMA.MY_TABLE a_field

    """

    func_name = 'map_list transform ' + event['_metadata']['event_type'] + ' ' + existing_column

    existing_column_value = event[existing_column]
    event[new_column] = None

    for k, v in mapping_dict_with_lists.items():
        if existing_column_value in v:
            event[new_column] = k

    if event[new_column] is not None:
        return event
    elif passthrough:
        event[new_column] = existing_column_value
    elif allow_nulls:
        return event
    else:
        raise Exception('Missing %s' % func_name)

    return event


def mark_for_delete(event):
    """ We created database triggers in our database to write all rows to a polymorphic deleted records table upon hard delete.
    We log the table_name, the time it was deleted, and the row_to_json
    This function creates a new row that looks like a soft delete came from the database

    :param event: A dictionary that includes the Alooma _metadata dictionary
    :return: A dictionary that looks like a soft deleted row
    
    Examples:

    .. code-block:: python

        # Example #1
        event = {'id': 1,
                'table_name': 'alooma_test',
                'primary_key': '123',
                'old_row_json': '{"id":6, "created_at":"2019-01-01"}',
                '_metadata': {'event_type': 'test'}}
        event = mark_for_delete(event)
        event = {'_metadata': {'event_type': 'alooma_test',
                                'table': 'alooma_test'},
                'created_at': '2019-01-01',
                'id': 6,
                'mark_for_delete': True}

    """

    event['_metadata']['event_type'] = event['table_name']
    event['mark_for_delete'] = True

    # Adds a column with the timestamp of when the row was hard deleted
    if 'created_at' in event:
        event['hard_deleted_at'] = event['created_at']
        del event['created_at']

    # Turns the old_row_json into the fields for the normal event
    for k, v in json.loads(event['old_row_json']).items():
        event[k] = v

    event['_metadata']['table'] = event['table_name']
    del event['old_row_json']
    del event['primary_key']
    del event['table_name']

    return event


def parse_list_of_json_and_concat(event, field_name, keep_original, field_to_keep):
    """ Iterates through a dictionary and creates a single field with a list of values from the field Output is similar to group_concat and listagg in SQL

    :param event: A dictionary
    :param field_name: The name of the field to extract data from
    :param keep_original: True to keep the original field.  False to delete the original field and only keep the parsed data.
    :param field_to_keep: The name of the field from within the dictionary
    :return: An altered dictionary

    Examples:

    .. code-block:: python

        # Example #1
        event = {'list_of_dicts': [{'key_to_concat': 123, 'key_to_ignore': 'abc'},
                                    {'key_to_concat': 456, 'key_to_ignore': 'def'},
                                    {'key_to_concat': 789, 'key_to_ignore': 'ghi'}]}
        event =  parse_list_of_json_and_concat(input_event, field_name='list_of_dicts', keep_original=True, field_to_keep='key_to_concat')
        event = {'list_of_dicts': [{'key_to_concat': 123, 'key_to_ignore': 'abc'},
                                    {'key_to_concat': 456, 'key_to_ignore': 'def'},
                                    {'key_to_concat': 789, 'key_to_ignore': 'ghi'}],
                'list_of_dicts_key_to_concats': [123, 456, 789]}

    """

    if field_name in event:
        temp = []
        for element in event[field_name]:
            for k, v in element.items():
                if k == field_to_keep:
                    field_name_underscore = field_name + '_'
                    if type(v) is dict:
                        temp.append(json.dumps(v))
                    else:
                        temp.append(v)

                    event[field_name_underscore + k + 's'] = temp

        if keep_original is False:
            del event[field_name]
    return event


def remove_duplicate_field(event, field_to_keep, field_to_discard):
    """ Remove a field when both fields are present

    :param event:  A dictionary
    :param field_to_keep: The field to keep if both keys are present and the value is not None
    :param field_to_discard: The field to discard if the field_to_keep is not None
    :return: An event with a single bundle ID

    Examples:

    .. code-block:: python

        # Example #1
        event = {'A_Field': 'another_value', 'a_field': 'a_value '}
        event = remove_duplicate_field(event, field_to_keep='a_field', field_to_discard='A_Field')
        event = {'a_field': 'a_value '}

    """

    if field_to_keep in event and field_to_discard in event:
        if event[field_to_keep] is not None:
            del event[field_to_discard]
        else:
            del event[field_to_keep]

    return event


def remove_outer_key(event, key_name):
    """ Removes the outer key from an event

    :param event: A dict with the entire event
    :param key_name: The key to remove from the dictionary
    :return: An event with the outer key for the specified dictionary removed

    Examples:

    .. code-block:: python

        # Example #1
        event = {'outer_dict': {'a_field': 'a_value ',
                                'another_field': 'another_value'}}
        event = remove_outer_key(event, key_name='outer_dict')
        event = {'a_field': 'a_value ',
                'another_field': 'another_value'}

    """

    # Removes the metrics key so we don't have every field prefixed with metrics_
    # if key_name in event and event[key_name] is dict:
    if key_name in event and isinstance(event[key_name], dict):
        event.update(event[key_name])
        del event[key_name]

    return event


def remove_starting_characters_from_keys(event, starting_characters, field_with_json=None):
    """ Removes the specified starting characters from all keys in an event

    :param event: A dict with the entire event
    :param starting_characters: The characters to remove from the beginning of the key
    :param field_with_json: A specific field with nested json from which to remove the characters from its keys
    :return: a modified event

    Examples:

    .. code-block:: python

        # Example #1
        event = {'_metadata': {},
                '$a_field': 'a_value',
                '$another_field': 'another_value'}
        event = remove_starting_characters_from_keys(event, starting_characters='$')
        event = {'_metadata': {},
                'another_field': 'another_value',
                'field': 'a_value'}

        # Example #2
        event = {'_metadata': {},
                'a_dict': {'$a_field': 'a_value',
                            '$another_field': 'another_value'},
                '$outer_field': 'some value'}
        event = remove_starting_characters_from_keys(event, starting_characters='$', field_with_json='a_dict')
        event = {'$outer_field': 'some value',
                '_metadata': {},
                'a_dict': {'a_field': 'a_value',
                            'another_field': 'another_value'}}

    """

    starting_character_length = len(starting_characters)

    if field_with_json is None:
        new_event = {}

        for k, v in event.items():
            if k.startswith(starting_characters):
                new_event[k[starting_character_length:]] = v
            else:
                new_event[k] = v

        new_event['_metadata'] = deepcopy(event['_metadata'])
        return new_event

    elif field_with_json is not None:
        new_dict = {}

        if field_with_json in event:
            for k, v in event[field_with_json].items():
                if k.startswith(starting_characters):
                    new_dict[k[starting_character_length:]] = v
                else:
                    new_dict[k] = v

            event[field_with_json] = new_dict
        return event


def remove_whitespace(event, field_or_field_list):
    """ Remove leading and trailing whitespace

    :param event: A dictionary
    :param field_or_field_list: A field or list of fields to trim the whitespace from
    :return: A trimmed string

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': '  should not have whitespace at ends    '}
        event = remove_whitespace(event, field_or_field_list='a_field')
        event = {'a_field': 'should not have whitespace at ends'}

        # Example #2
        event = {'a_field': '  should not have whitespace at ends    ',
                'another_field': '  also should not have whitespace at ends    '}
        event = remove_whitespace(event, field_or_field_list=['a_field', 'another_field'])
        event = {'a_field': 'should not have whitespace at ends',
                'another_field': 'also should not have whitespace at ends'}


    """

    if type(field_or_field_list) is not list:
        field_or_field_list = [field_or_field_list]

    for field in field_or_field_list:
        if field in event and isinstance(event[field], str):
            event[field] = event[field].strip()

    return event


def rename_fields(event, field_dict):
    """ Renames fields from the key to value
    :param event: A dict with the entire event
    :param field_dict: A dict with the rename mapping with the key as the existing field and the value as the new field
    :return: An altered event with the renamed fields

    Examples:

    .. code-block:: python

        # Example #1
        event = {'a_field': 'a_value',
                'another_field': 'another_value',
                'no_change': 'same'}
        event = rename_fields(event, field_dict={'a_field': 'field_one', 'another_field': 'field_two'})
        event = {'field_one': 'a_value',
                'field_two': 'another_value',
                'no_change': 'same'}

    """

    new_event = {}

    for k, v in event.items():
        if k in field_dict:
            if field_dict[k] not in new_event:
                new_event[field_dict[k]] = event[k]
            else:
                raise Exception('Key {key_name} already in event'.format(key_name=field_dict[k]))
        else:
            new_event[k] = event[k]

    if '_metadata' in event:
        new_event['_metadata'] = deepcopy(event['_metadata'])

    return new_event


def split_event_to_multiple_events(event, table_name_list):
    """ Splits events into a list of events with a schema_name.table_name

    :param event: A dict with a single event
    :param table_name_list: The table names for the new events
    :return: A list of the new events.  If an event has already been split it will not re-split and returns the original event.

    Examples:

    .. code-block:: python

        # Example #1
        event = {'_metadata': {'@uuid': '123-abc', 'event_type':
                                'my_schema.my_table'},
                'a_field': 'a_value'}
        event = split_event_to_multiple_events(event, table_name_list=['table_one', 'table_two'])
        # A parent UUID is added in Alooma when events are split.  Local testing won't add a parent UUID.
        event = [{'_metadata': {'@parent_uuid': '123-abc',
                                '@uuid': '456-def',
                                'event_type': 'my_schema.table_one'},
                'a_field': 'a_value'},
                {'_metadata': {'@uuid': '123-abc',
                                '@uuid': '789-ghi',
                                'event_type': 'my_schema.table_two'},
                'a_field': 'a_value'}]

        # Example #2
        event = {'_metadata': {'@parent_uuid': '123-abc',
                                '@uuid': '456-def',
                                'event_type': 'my_schema.table_one'},
                'a_field': 'a_value'}
        event = split_event_to_multiple_events(event, table_name_list=['table_one', 'table_two'])
        # If the event has a parent_uuid it will not be re-split
        event = {'_metadata': {'@parent_uuid': '123-abc',
                                '@uuid': '456-def',
                                'event_type': 'my_schema.table_one'},
                'a_field': 'a_value'}

    """

    # Avoids re-splitting events
    if '@parent_uuid' in event['_metadata'] and len(event['_metadata']['@parent_uuid']) > 0:
        return event

    if '.' not in event['_metadata']['event_type']:
        raise Exception('Only fully qualified events can be split. Event type must be schema_name.table_name')

    event_list = []

    for table_name in table_name_list:
        new_event = {}

        for k, v in event.items():
            if k != "_metatdata":
                new_event[k] = v

        new_event['_metadata'] = deepcopy(event['_metadata'])
        new_event['_metadata']['event_type'] = event['_metadata']['event_type'].split('.')[0] + '.' + table_name
        event_list.append(new_event)

    return event_list


def split_field_list_to_multiple_events(event, fields_to_split, add_counter, counter_name, reverse=False):
    """ Take an event that has columns that are lists (of same length) and break into multiple rows

    :param event: A dictionary
    :param fields_to_split: The field with to split
    :param add_counter: True to add a counter field to the event
    :param counter_name: The name of the counter field
    :param reverse: True to start the counter in reverse
    :return: A list of events

    Examples:

    .. code-block:: python

        # Example #1
        event = {'id': 1,
                'names': ['first', 'second'],
                '_metadata': {'uuid': '1a'}}
        event =  split_field_list_to_multiple_events(event=input, fields_to_split=['names'], add_counter=True, counter_name='counter', reverse=False)
        event = [{'id': 1,
                   'name': 'first',
                   'counter': 1,
                   '_metadata': {'@parent_uuid': '1a',
                                '@uuid': '456-def'}},
                  {'id': 1,
                   'name': 'second',
                   'counter': 2,
                   '_metadata': {'@parent_uuid': '1a',
                                '@uuid': '789-ghi'}}]

    """

    # Avoids re-splitting events
    if '@parent_uuid' in event['_metadata'] and len(event['_metadata']['@parent_uuid']) > 0:
        return event

    event_list = []
    number_events = len(event[fields_to_split[0]])
    for i in range(number_events):
        new_event = {}
        for k, v in event.items():
            if k in fields_to_split:
                if reverse:
                    v = v[::-1]
                if i + 1 > len(v):
                    new_event[convert_string_to_snake_case(k[:-1])] = None
                else:
                    new_event[convert_string_to_snake_case(k[:-1])] = v[i]
            elif k != "_metadata":
                new_event[convert_string_to_snake_case(k)] = v

        new_event['_metadata'] = deepcopy(event['_metadata'])
        if add_counter:
            new_event[counter_name] = i + 1
        event_list.append(new_event)

    return event_list


def whitelist_or_blacklist_columns(event, field_list, white_or_black_list='whitelist'):
    """ Allows you to remove a list of fields (blacklist) or limit an event to a list of fields (whitelist)

    :param event: A dictionary
    :param field_list: A list of fields to keep or remove
    :param white_or_black_list: whitelist = Only let a particular list of columns through the event and remove other columns. blacklist = Don't allow a particular list of columns through. Leave all other columns
    :return: An altered dictionary

    Examples:

    .. code-block:: python

        # Example #1
        event = {'_metadata': {},
                'keep_me': 'i stay',
                'keep_me_too': 'i stay too',
                'remove_me': 'im gone'}
        event = whitelist_or_blacklist_columns(event, field_list=['keep_me', 'keep_me_too'], white_or_black_list='whitelist')
        event = {'_metadata': {},
                'keep_me': 'i stay',
                'keep_me_too': 'i stay too'}

        # Example #2
        event = {'_metadata': {},
                'keep_me': 'i stay',
                'keep_me_too': 'i stay too',
                'remove_me': 'im gone'}
        event = whitelist_or_blacklist_columns(event, field_list=['remove_me'], white_or_black_list='blacklist')
        event = {'_metadata': {},
                'keep_me': 'i stay',
                'keep_me_too': 'i stay too'}

    """

    field_list = [f.lower() for f in field_list]

    new_event = {}

    if white_or_black_list == 'whitelist':
        for k, v in event.items():
            if k != "_metadata" and k.lower() in field_list:
                new_event[k] = v

    elif white_or_black_list == 'blacklist':
        for k, v in event.items():
            if k != "_metadata" and k.lower() not in field_list:
                new_event[k] = v

    new_event['_metadata'] = deepcopy(event['_metadata'])

    return new_event



import pandas as pd
from tabulate import tabulate
import pprint
import requests


class Events:

    def __init__(self, api):
        """ Gets information from Alooma on events and prints detailed information about events.

        :param api: api authentication using the Alooma package
        """
        self.api = api

    @staticmethod
    def print_sorted_list(lst):
        """
        :param lst: list Takes a list and prints a sorted list
        """
        print('\n'.join(sorted(lst)))

    def get_all_events(self):
        """ Gets the information for all events from Alooma

        :return: A list of dictionaries with information for each event
        """
        return self.api.get_event_types()

    def view_events(self, single_input=None, single_event=None, print_format='table', table_limit=None, pprint_indent=2, pprint_width=250, pprint_depth=5):
        """ Prints a data from with the event info

        :param single_input: The name of a specific input to filter the results
        :param single_event: The name of a specific event to filter the results
        :param print_format: Specify 'table' to print event info as tables or 'json' to print as dictionaries
        :param table_limit: Limit the number of events printed in the dataframe
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        :param pprint_depth: The depth value to pprint dictionaries
        :return: The event data in the print_format of a dataframe or a list of dictionaries
        """

        events = self.get_all_events()

        if single_event:
            events = [e for e in events if e["name"] == single_event]

        if single_input:
            events = [e for e in events if e["origInputLabel"] == single_input]

        if print_format == 'table':
            df = pd.DataFrame(events)
            df = df.reindex(['origInputLabel', 'name', 'state', 'stats', 'mappingMode', 'autoMappingError',
                                  'usingDefaultMappingMode', 'schemaUrls', 'consolidation'], axis=1)
            df.sort_values(by=['origInputLabel', 'name'], ascending=True, inplace=True)
            print(tabulate(df.head(n=table_limit), headers='keys', tablefmt='psql', showindex=False))
            return df

        elif print_format == 'json':
            for e in events:
                pprint.pprint(e, indent=pprint_indent, width=pprint_width, depth=pprint_depth)
            return events

    def list_events(self, input_labels='all', print_lst=False, add_quotes_commas=False):
        """ Prints and/or returns a list of event names

        :param input: string 'all' for all events or specify an input label for a specific events
        :param print_lst: boolean True to print the list to the console
        :param add_quotes_commas: boolean True to print the list to the console with single quotes around strings and commas after each item
        :return: list of events by name
        """
        events = self.get_all_events()

        if input_labels != 'all':
            events = [e["name"] for e in events if input_labels == e["origInputLabel"]]
        else:
            events = [e["name"] for e in events]

        if print_lst:
            if not add_quotes_commas:
                self.print_sorted_list(events)
            else:
                events = ["'" + e + "'," for e in events]
                self.print_sorted_list(events)

        return events

    def delete_event(self, event_name):
        """ Deletes an event from the mapper and prints the event_name and API response

        :param event_name: The name of the event to delete
        :return: None
        """

        res = self.api.delete_event_type(event_name)
        print("Deleted event", event_name)

        return res




import pprint


class DictionaryDifferences:

    def __init__(self, old_dictionary, new_dictionary, pprint_indent=2, pprint_width=250, pprint_depth=5):
        """ Takes 2 Dictionaries and has methods to print the out

        :param old_dictionary:
        :param new_dictionary:
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        :param pprint_depth: The depth value to pprint dictionaries
        """

        self.old_dictionary = old_dictionary
        self.new_dictionary = new_dictionary
        self.pprint_indent = pprint_indent
        self.pprint_width = pprint_width
        self.pprint_depth = pprint_depth

    def show_dictionary_all(self):
        """ Shows the before and after transformation for entire dictionaries

        :return: None
        """

        if self.old_dictionary is not None and self.new_dictionary is not None:
            print('\n' * 2)
            print('ORIGINAL DICTIONARY'.center(200, '-'))
            pprint.pprint(self.old_dictionary, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)
            print('\n' * 2)
            print('ALTERED DICTIONARY'.center(200, '-'))
            pprint.pprint(self.old_dictionary, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)
        elif self.old_dictionary is not None:
            print('\n' * 2)
            print('DICTIONARY'.center(200, '-'))
            pprint.pprint(self.old_dictionary, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)
        elif self.new_dictionary is not None:
            print('\n' * 2)
            print('DICTIONARY'.center(200, '-'))
            pprint.pprint(self.new_dictionary, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

    def show_dictionary_differences(self, show_matching=True, show_changed=True, show_removed=True, show_added=True):
        """ Shows the matching, changed, removed, and added dictionary keys and values

        :return: None
        """

        removed_key_and_value = {}
        added_key_and_value = {}
        changed_values = {}
        matching_keys_and_values = {}

        for key, value in self.old_dictionary.items():
            if key in self.new_dictionary and value == self.new_dictionary[key]:
                matching_keys_and_values[key] = value
            if key in self.new_dictionary and value != self.new_dictionary[key]:
                changed_values[key] = {'old_value': value, 'new_value': self.new_dictionary[key]}
            if key not in self.new_dictionary:
                removed_key_and_value[key] = value

        for key, value in self.new_dictionary.items():
            if key not in self.old_dictionary:
                added_key_and_value[key] = value

        if show_matching:
            print('\n' * 2)
            print('MATCHING KEYS AND VALUES'.center(200, '-'))
            pprint.pprint(matching_keys_and_values, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        if show_changed:
            print('\n' * 2)
            print('CHANGED VALUES'.center(200, '-'))
            pprint.pprint(changed_values, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        if show_removed:
            print('\n' * 2)
            print('REMOVED KEYS VALUES'.center(200, '-'))
            pprint.pprint(removed_key_and_value, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        if show_added:
            print('\n' * 2)
            print('ADDED KEYS AND VALUES'.center(200, '-'))
            pprint.pprint(added_key_and_value, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)





import pprint
import requests
import pandas as pd
from tabulate import tabulate
from collections import OrderedDict
import copy


class Consolidations:

    def __init__(self, api, preview_changes=True, apply_changes=False, pprint_indent=2, pprint_width=250, pprint_depth=5):
        """Consolidations are Alooma's way of making sure only the most recent row for each table appears in the main table.
         Consolidation management the v2 API.  The class is initiated with the following variables:

        :param api: The Alooma API client authentication
        :param preview_changes: Prints the consolidation or consolidations changes if True
        :param apply_changes: Executes the consolidation changes if True
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        :param pprint_depth: The depth value to pprint dictionaries

        """

        self.api = api
        self.preview_changes = preview_changes
        self.apply_changes = apply_changes
        self.pprint_indent = pprint_indent
        self.pprint_width = pprint_width
        self.pprint_depth = pprint_depth

    def get_scheduled_queries(self):
        """ Gets the list of scheduled queries from Alooma

        :return: The JSON response with all of the scheduled queries
        """

        # Consolidation V2
        CONSOLIDATION_V2 = 'v2/consolidation'
        CONSOLIDATION_STATE_V2 = 'v2/consolidation/{query_id}'
        CONSOLIDATION_RUN_V2 = 'v2/consolidation/{query_id}/run'

        url = self.api.rest_url + 'v2/consolidation'
        res = self.api._Client__send_request(requests.get, url)

        return res.json()

    def scheduled_query_table(self):
        """ Retrieves a list of the consolidation queries running in Alooma. Prints the information in a dataframe if the class preview_changes flag is True.

        :return: A dataframe with the information on the scheduled queries
        """
        scheduled_queries = self.get_scheduled_queries()

        orderedDictList = []
        for query in scheduled_queries:

            orderedDictList.append(OrderedDict([('id', query['id']),
                                                ('event_name', query['event_type']),
                                                ('error_message', query['error_message']),
                                                ('is_active', query['is_active']),
                                                ('is_running', query['is_running']),
                                                ('schedule', query['schedule']),
                                                ('last_success_time', query['last_success_time']),
                                                ('last_run_start_time', query['last_run_start_time']),
                                                ('last_run_end_time', query['last_run_end_time']),
                                                ('last_run_return_code', query['last_run_return_code']),
                                                ('start_time', query['start_time']),
                                                ('next_run_time', query['next_run_time']),
                                                ('generated_by', query['generated_by']),
                                                ('query_type', query['query_type']),
                                                ('service_type', query['service_type']),
                                                ('time_limit', query['time_limit']),
                                                ('name', query['name']),
                                                ('service_type', query['service_type']),
                                                ('enqueue_time', query['enqueue_time']),
                                                ('docker_tag', query['docker_tag']),
                                                ('docker_img', query['docker_img']),
                                                ('docker_cmd', query['docker_cmd'])
                                               ]))

        df = pd.DataFrame(orderedDictList).sort_values('event_name')

        if self.preview_changes:
            print(tabulate(df, headers='keys', tablefmt='psql', showindex=False))

        return df

    def create_consolidation(self, event_name, cron_schedule='*/15 * * * *', custom_variables=False):
        """ Create Consolidation Given Configuration

            :param event_name: The event type for which to create the consolidation
            :param cron_schedule: The cron schedule on which to run the schedule query. Default is 15 minutes.
            :param custom_variables: custom variables to add to consolidation. For consolidations from database tables set custom variables to None. For consolidations for all other tables the recommended customer variable settings are custom_variables = {"all_ri_fields": ["_metadata.@timestamp"], "ri_field": "_metadata.@timestamp", "ri_column": "_metadata.@timestamp"}

        """

        configuration = {
            "event_type": event_name,
            "run_at": cron_schedule,
            "query_type": 'incremental'
        }

        # For non-database tables use the metadata fields for custom variables
        if custom_variables:
            configuration["custom_variables"] = {"all_ri_fields": ["_metadata.@timestamp"], "ri_field": "_metadata.@timestamp", "ri_column": "_metadata.@timestamp"}

        if "deployment_name" not in configuration:
            deployment_name = self.api.get_deployment_info()['deploymentName']
            configuration["deployment_name"] = deployment_name

        if self.preview_changes:
            pprint.pprint(configuration, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        if self.apply_changes:
            url = self.api.rest_url + 'v2/consolidation'
            res = self.api._Client__send_request(requests.post, url, json=configuration)
            pprint.pprint(res.json(), indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        return configuration

    def get_scheduled_query_for_event(self, event_name):
        """ Gets the scheduled query for a specific event

        :param event_name: The name of the event for which to retrieve the schedule query
        :return: Returns the schedule query
        """

        schqueries = self.get_scheduled_queries()
        schquery = [c for c in schqueries if c['event_type'] == event_name]

        return schquery[0]

    def view_schedule_query_for_event(self, event_name, show_configuration_only=True):
        """ Prints the scheduled query for an event when

        :param event_name: The event to view a scheduled query for
        :param show_configuration_only: Print the configuration only without the query
        :return: The scheduled query
        """

        try:
            schquery = self.get_scheduled_query_for_event(event_name)
        except IndexError:
            print ("No Consolidation for", event_name)

        if self.preview_changes:
            if show_configuration_only:
                consolidation_without_details = copy.deepcopy(schquery)
                consolidation_without_details.pop('consolidation_query', None)
                pprint.pprint(consolidation_without_details, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)
            else:
                pprint.pprint(schquery, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        return schquery

    def remove_scheduled_query(self, schquery):
        """ Removes the scheduled query for a specific ID

        :param schquery: the ID of a schedule query to remove
        :return: None
        """
        url = self.api.rest_url + 'v2/consolidation/' + str(schquery['id'])

        if self.apply_changes:
            res = self.api._Client__send_request(requests.delete, url)
            print(res)

        return None

    def remove_scheduled_query_for_event(self, event_name, show_configuration_only=True):
        """ Removes the scheuled query for an event

        :param event_name: The name of the event for which to remove the consolidation
        :param print_consolidations: print the current consolidation
        :param show_configuration_only: Print the configuration only without the query
        :param apply_change: apply the remove consolidation
        :return:
        """
        schquery = self.get_scheduled_query_for_event(event_name)

        if self.preview_changes:
            if show_configuration_only:
                consolidation_without_details = copy.deepcopy(schquery)
                consolidation_without_details.pop('consolidation_query', None)
                pprint.pprint(consolidation_without_details, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)
            else:
                pprint.pprint(schquery, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        if self.apply_changes:
            self.remove_scheduled_query(schquery)

        return None

    def change_scheduled_query_frequency_for_event(self, event_name, cron_schedule='*/15 * * * *'):
        """ Change the frequency of the scheduled query

        :param event_name: he name of the event for which to remove the consolidation
        :param cron_schedule: The new cron schedule to set for the event
        :return: None
        """

        schquery = self.get_scheduled_query_for_event(event_name)

        if self.preview_changes:
            print("OLD FREQUENCY", schquery['configuration']['run_at'])

        schquery['configuration']['run_at'] = cron_schedule

        if self.preview_changes:
            print ("NEW FREQUENCY", schquery['configuration']['run_at'])

        if self.apply_changes:
            res = self.remove_scheduled_query(schquery)
            print("REMOVING OLD QUERY", res)
            res = self.create_consolidation_v2(self.api, schquery['configuration'])
            print("RE-CREATING QUERY WITH NEW SCHEDULE", res)

        return None



import pprint
import copy
import os

from .dictionary_difference import DictionaryDifferences
from .samples import Samples


class TransformationTest:

    def __init__(self, api=None, code_package=None, preview_full_events=True, preview_difference_dicts=True, local_or_api='local', pprint_indent=2, pprint_width=250, pprint_depth=5):
        """ Tests the Alooma events

        :param api: api authentication using the Alooma package
        :param preview: True to print the transformations for visual inspection
        :param code_package: Your specific transformation code to use as a package. This is needed for local testing
        :param local_or_api: 'local' to test on local code or 'api' to test on code deployed to Alooma
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        """

        self.api = api
        self.preview_full_events = preview_full_events
        self.preview_difference_dicts = preview_difference_dicts
        self.code_package = code_package
        self.local_or_api = local_or_api
        self.pprint_indent = pprint_indent
        self.pprint_width = pprint_width
        self.pprint_depth = pprint_depth

    def test_single_event(self, sample):
        """ Tests a single event using the local code and pretty prints the before and after dictionaries to the console

        :param sample: A single sample event dictionary
        :return: The output dictionary
        """

        # Copy the original event for diffs to work
        original_event = copy.deepcopy(sample)

        # Transform the even in the API or locally
        if self.local_or_api == 'local':
            transformed_event = self.code_package.transform(sample)
        elif self.local_or_api == 'api':
            output = self.api.test_code_engine_code(sample)
            transformed_event = output['result']

        # Print the before and after full events
        if self.preview_full_events:
            print('INPUT EVENT'.center(200, '-'))
            pprint.pprint(original_event, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)
            print('\n' * 2)
            print('OUTPUT EVENT'.center(200, '-'))
            pprint.pprint(transformed_event, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)
            if self.local_or_api == 'api':
                if len(output['notifications']) > 0:
                    print('\n' * 2)
                    print('API NOTIFICATIONS'.ljust(200, '-'))
                    print(output['notifications'])

        # Print dictionaries categorizing the differences in the transformation
        if self.preview_difference_dicts:
            DD = DictionaryDifferences(old_dictionary=original_event, new_dictionary=transformed_event)
            DD.show_dictionary_differences()

        if transformed_event is not None and 'errorMessage' in transformed_event:
            self.fail(
                'Failed on event "%s"\n%s' % (
                    sample[sample],
                    transformed_event['errorMessage']))

        return transformed_event

    def loop_through_events_to_test(self, sample_event_list):
        """ Loops through the event files to test each sample

        :param sample_event_list: A list of dictionaries to test. The expected format from Alooma is a dictionary as formatted below
        :return: None

        Alooma samples are in a list of dictionaries with the sample inside of a key of sample:

        .. code-block:: python

            [{'sample': {}
            {'sample': {}
            {'sample': {}]

        """

        for event in sample_event_list:
            self.test_single_event(sample=event['sample'])

        return None

    def test_all_events(self, sample_event_directory, file_name, file_prefix='input'):
        """ Tests all events from a specified file or all the saved input sample.

        :param sample_event_directory: The name of the directory where sample files are saved.
        :param file_name: The name of a specific file to test. A file should be a list with dictionaries.
        :param file_prefix: Specific a file prefix to test all the events in files with a similar name.  The defaul will test all the samples for all inputs that have been saved using the Samples.save_alooma_samples_to_files function.
        :return: True if the samples should be printed to the console
        """

        S = Samples(api=self.api, sample_event_directory=sample_event_directory)

        if file_name is None:
            for file in os.listdir(sample_event_directory):
                if file.endswith(".json") and file.startswith(file_prefix):
                    file_path = (os.path.join(sample_event_directory, file))
                    sample_event_list = S.get_samples_from_file(file_path)
                    self.loop_through_events_to_test(sample_event_list)

        else:
            file_path = (os.path.join(sample_event_directory, file_name))
            sample_event_list = S.get_samples_from_file(file_path)
            self.loop_through_events_to_test(sample_event_list)

        return None




from .consolidations import Consolidations
from .column_ddl import ColumnDDL
from .dictionary_difference import DictionaryDifferences
from .events import Events
from .inputs import Inputs
from .mappings import Mappings
from .samples import Samples
from .secrets import Secrets
from .transformation_test import TransformationTest
from .transformation_functions import (add_column_based_on_null,
                                       add_columns_with_default,
                                       add_composite_key,
                                       add_duplicate_fields,
                                       add_suffix,
                                       convert_all_event_fields_to_snake_case,
                                       convert_dictionary_fields_to_string,
                                       convert_empty_value_to_none,
                                       convert_event_type_case,
                                       convert_null_to_zero,
                                       convert_spaces_and_special_characters_to_underscore,
                                       convert_string_to_snake_case,
                                       convert_values_to_none,
                                       flatten_json,
                                       flatten_json_1_level,
                                       map_key_in_dictionary_to_value,
                                       map_value_in_list_to_dictionary_key,
                                       mark_for_delete,
                                       parse_list_of_json_and_concat,
                                       remove_duplicate_field,
                                       remove_outer_key,
                                       remove_starting_characters_from_keys,
                                       remove_whitespace,
                                       rename_fields,
                                       split_event_to_multiple_events,
                                       split_field_list_to_multiple_events,
                                       whitelist_or_blacklist_columns)



import pprint
import re
import copy
import requests
from .dictionary_difference import DictionaryDifferences


class Mappings:

    def __init__(
            self, api, event_name, preview_full=True, preview_changes=True,
            apply_changes=False, pprint_indent=2, pprint_width=250,
            pprint_depth=5):
        """View and change Alooma mappings. Mappings are the conversion of a dictionary to a flattened table structure.
        The class is initiated with the following variables:

        :param api: The Alooma API client authentication
        :param event_name: The name of the event to view or change settings for
        :param preview_full: Prints the mapping or mapping changes if True. The default is True.
        :param preview_changes: Prints the changes in the mapping by category if True
        :param apply_changes: Executes the mapping changes if True
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        :param pprint_depth: The depth value to pprint dictionaries

        """
        self.api = api
        self.event_name = event_name
        self.preview_full = preview_full
        self.preview_changes = preview_changes
        self.apply_changes = apply_changes
        self.pprint_indent = pprint_indent
        self.pprint_width = pprint_width
        self.pprint_depth = pprint_depth

    def get_mapping_for_event(self):
        """ Gets the mapping for the event name that the class was initialized with

        :return: a dictionary with the mapping
        """
        return self.api.get_mapping(self.event_name)

    def view_mapping(self, view_field_mappings=False):
        """ Gets the mapping for an event and allows printing with or with the field details

        :param view_field_mappings: Hides the field details when set to false. Useful for quick view of consolidation and mapping details without all the fields
        :return: Returns the mapping dictionary the user viewed

        Sample Mapping printed with view_field_mappings=False.  The mapping['fields'] key with the list of fields IS NOT printed.

        .. code-block:: python

          {'autoMappingError': None,
           'consolidation': {'consolidatedSchema': 'MY_SCHEMA',
                            'consolidatedTableName': 'MY_TABLE',
                            'consolidationKeys': ['ID'],
                            'viewSchema': None},
           'inputObjects': {'12345-asdfg': ['98765-zxcvb']},
           'mapping': {'isDiscarded': False,
                       'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
                       'outputId': 'a1s2d3-f4g5h6',
                       'readOnly': False,
                       'schema': 'MY_SCHEMA',
                       'tableName': 'MY_TABLE_LOG'},
            'mappingMode': 'AUTO_MAP',
            'name': 'MY_SCHEMA.MY_TABLE',
            'origInputLabel': 'production_database',
            'schemaUrls': ['schema?id=12345-asdfg&schema_object=my_table',
                        'schema?id=d=12345-asdfg&sschema_object=deleted_rows'],
            'state': 'MAPPED',
            'usingDefaultMappingMode': False}



        Sample Mapping printed with view_field_mappings=True. The mapping['fields'] key with the list of fields IS printed.

        .. code-block:: python

          {'autoMappingError': None,
           'consolidation': {'consolidatedSchema': 'MY_SCHEMA',
                            'consolidatedTableName': 'MY_TABLE',
                            'consolidationKeys': ['ID'],
                            'viewSchema': None},
            'fields': [ {'fieldName': 'id',
                        'fields': [],
                        'mapping': {'columnName': 'ID',
                        'columnType': {'nonNull': True,
                                        'precision': 38,
                                        'scale': 0,
                                        'type': 'NUMERIC'},
                                        'isDiscarded': False,
                                        'machineGenerated': False,
                                        'subFields': None}},
                        {'fieldName': 'name',
                        'fields': [],
                        'mapping': {'columnName': 'NAME',
                        'columnType': {'length': 16777216,
                                        'nonNull': False,
                                        'truncate': False,
                                        'type': 'VARCHAR'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}
                        ],
           'inputObjects': {'12345-asdfg': ['98765-zxcvb']},
           'mapping': {'isDiscarded': False,
                       'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
                       'outputId': 'a1s2d3-f4g5h6',
                       'readOnly': False,
                       'schema': 'MY_SCHEMA',
                       'tableName': 'MY_TABLE_LOG'},
            'mappingMode': 'AUTO_MAP',
            'name': 'MY_SCHEMA.MY_TABLE',
            'origInputLabel': 'production_database',
            'schemaUrls': ['schema?id=12345-asdfg&schema_object=my_table',
                        'schema?id=d=12345-asdfg&sschema_object=deleted_rows'],
            'state': 'MAPPED',
            'usingDefaultMappingMode': False}


        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping to remove the fields
        new_mapping.pop('fields', None)

        if self.preview_changes and not self.preview_full:
            raise Exception(
                'You must specify preview_changes=True to see the mapping. This function does not change the mapping and preview_changes is ignored')

        # Print the dictionary with or without the fields
        if view_field_mappings:
            DD = DictionaryDifferences(
                old_dictionary=mapping,
                new_dictionary=None,
                pprint_indent=self.pprint_indent,
                pprint_width=self.pprint_width,
                pprint_depth=self.pprint_depth)
            DD.show_dictionary_all()
            return mapping
        else:
            DD = DictionaryDifferences(
                old_dictionary=None,
                new_dictionary=new_mapping,
                pprint_indent=self.pprint_indent,
                pprint_width=self.pprint_width,
                pprint_depth=self.pprint_depth)
            DD.show_dictionary_all()
            return new_mapping

    def change_mapping_mode(self, new_mapping_mode='STRICT'):
        """ Change the mapping mode. Alooma has 3 modes of AUTO_MAP, STRICT, and FLEXIBLE. We only use AUTO_MAP or STRICT

        :param new_mapping_mode: The new mapping mode to set: AUTO_MAP, STRICT, and FLEXIBLE
        :return: The altered mapping

        The mapping settings keys and sample values

        .. code-block:: python

            {'autoMappingError': None,
            'mappingMode': None,
            'usingDefaultMappingMode': True}

        The mapping mode with the specified new mappingMode to alter

        .. code-block:: python

            {'autoMappingError': None,
            'mappingMode': 'AUTO_MAP',
            'usingDefaultMappingMode': True}

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        new_mapping['mappingMode'] = new_mapping_mode

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="NEW MAPPING MODE SET")

        return new_mapping

    def change_mapping_consolidation_settings(
            self, consolidation_schema, consolidation_table_name,
            consolidation_keys):
        """ Updates the consolidation information for an event

        :param consolidation_schema: The schema of consolidated table
        :param consolidation_table_name: The name of the consolidated table
        :param consolidation_keys: The consolidation keys (primary key) for the table. Takes a single field or a list
        :return: The altered mapping

        The old consolidation settings

        .. code-block:: python

            {'consolidatedSchema': 'MY_SCHEMA_TEMP',
            'consolidatedTableName': 'NY_TABLE_ITEMS',
            'consolidationKeys': ['IDENTIFIER'],
            'viewSchema': None}

        The new consolidation settings to apply

        .. code-block:: python

            {'consolidatedSchema': 'MY_SCHEMA',
            'consolidatedTableName': 'MY_TABLE',
            'consolidationKeys': ['ID'],
            'viewSchema': None}

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        if not isinstance(consolidation_keys, list):
            key_list = []
            key_list.append(consolidation_keys)

        new_mapping['consolidation']['consolidatedSchema'] = consolidation_schema
        new_mapping['consolidation']['consolidatedTableName'] = consolidation_table_name
        new_mapping['consolidation']['consolidationKeys'] = key_list

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="UPDATED CONSOLIDATION")

        return new_mapping

    def change_mapping_consolidation_key(self, new_consolidation_key):
        """ Change the conolidation key only

        :param new_consolidation_key: The new consolidation key. This is the primary key for the table.
        :return: The altered mapping

        The consolidation key is set to the field name ID when tables are auto-mapped

        .. code-block:: python

            {'consolidatedSchema': 'MY_SCHEMA',
            'consolidatedTableName': 'MY_TABLE',
            'consolidationKeys': ['ID'],
            'viewSchema': None}

        This example will change the key to a field named IDENTIFIER

        .. code-block:: python

            {'consolidatedSchema': 'MY_SCHEMA',
            'consolidatedTableName': 'MY_TABLE',
            'consolidationKeys': ['IDENTIFIER'],
            'viewSchema': None}
        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        new_mapping['consolidation']['consolidationKeys'] = new_consolidation_key

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="CONSOLIDATION KEY SET")

        return new_mapping

    def change_mapping_to_use_log(self):
        """ Changes the mapping from using the consolidated table to the log table. Used when adjusting manual mappings

        :param event_name: The event name of the mapping to alter
        :return: The altered mapping

        Events without a _log do not have consolidated table information

        .. code-block:: python

             'mapping': {'isDiscarded': False,
              'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
              'outputId': 'a1s2d3-f4g5h6',
              'readOnly': False,
              'schema': 'MY_SCHEMA',
              'tableName': 'MY_TABLE'}

        This example will change the key to a field named IDENTIFIER

        .. code-block:: python

            'mapping': {'isDiscarded': False,
              'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
              'outputId': 'a1s2d3-f4g5h6',
              'readOnly': False,
              'schema': 'MY_SCHEMA',
              'tableName': 'MY_TABLE_LOG'}

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        table_name_with_log = mapping['mapping']['tableName'] + '_LOG'
        new_mapping['mapping']['tableName'] = table_name_with_log

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="LOG TABLE SET")

        return new_mapping

    def change_mapping_for_manual_consolidation_creation(
            self, consolidation_schema, consolidation_table_name,
            consolidation_keys, case='UPPER'):
        """ Updates the mapping after creating the log table manually

        :param consolidation_schema: The schema of consolidated table
        :param consolidation_table_name: The name of the consolidated table
        :param consolidation_keys: The consolidation keys (primary key) for the table. Takes a single field or a list
        :param case: UPPER if your events are MY_SCHEMA.MY_EVENT and LOWER if the events are my_schema_my_event
        :return: The altered mapping

        First adjust the tables in the data warehouse.

        .. code-block:: python

            ALTER TABLE MY_TABLE RENAME TO MY_TABLE_LOG;
            CREATE TABLE MY_TABLE LIKE MY_TABLE_LOG;

        Then run change_mapping_for_manual_consolidation_creation  to change the mapping to insert data into the new log table. This is the old mapping:

        .. code-block:: python

            'consolidation': {'consolidatedSchema': None,
                            'consolidatedTableName': None,
                            'consolidationKeys': ,
                            'viewSchema': None},
           'mapping': {'isDiscarded': False,
                       'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
                       'outputId': 'a1s2d3-f4g5h6',
                       'readOnly': False,
                       'schema': 'MY_SCHEMA',
                       'tableName': 'MY_TABLE'},

        These are the changes that will be applied with the new mapping

        .. code-block:: python

            'consolidation': {'consolidatedSchema': 'MY_SCHEMA',
                            'consolidatedTableName': 'MY_TABLE',
                            'consolidationKeys': ['ID_FIELD'],
                            'viewSchema': None},
           'mapping': {'isDiscarded': False,
                       'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
                       'outputId': 'a1s2d3-f4g5h6',
                       'readOnly': False,
                       'schema': 'MY_SCHEMA',
                       'tableName': 'MY_TABLE_LOG'},

        Lastly add consolidation queries to combine the new data with the existing data using :func:`~managealooma.consolidations.Consolidations.create_consolidation`

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        if not isinstance(consolidation_keys, list):
            key_list = []
            key_list.append(consolidation_keys)

        # Enforce casing
        if case == 'UPPER':
            key_list = [k.upper() for k in key_list]
            consolidation_schema = consolidation_schema.upper()
            consolidation_table_name = consolidation_table_name.upper()
            mapping_table_name = mapping['mapping']['tableName'].upper(
            ) + '_LOG'
        elif case.lower() == 'lower':
            key_list = [k.lower() for k in key_list]
            consolidation_schema = consolidation_schema.lower()
            consolidation_table_name = consolidation_table_name.lower()
            mapping_table_name = mapping['mapping']['tableName'].lower(
            ) + '_log'

        # Set the new fields with the default case of UPPER
        new_mapping['consolidation']['consolidatedSchema'] = consolidation_schema.upper()
        new_mapping['consolidation']['consolidatedTableName'] = consolidation_table_name.upper()
        new_mapping['consolidation']['consolidationKeys'] = key_list
        new_mapping['mapping']['tableName'] = mapping_table_name

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="UPDATED CONSOLIDATION")

        return new_mapping

    def delete_field_from_mapping(self, field_name):
        """ Gets a mapping, deletes a field from the mapping, and resets the mapping table statistics.

        :param event_name: the name of event for which to make the change
        :param field_name: the field name that should be deleted
        :return: The altered mapping

        If print_changes is specified the details for the field to remove is printed

        .. code-block:: python

           {'fieldName': 'name',
                        'fields': [],
                        'mapping': {'columnName': 'NAME',
                        'columnType': {'length': 16777216,
                                        'nonNull': False,
                                        'truncate': False,
                                        'type': 'VARCHAR'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}



        Then the entire new mapping is printed. The 'name' field is not longer in the mapping

        .. code-block:: python

          {'autoMappingError': None,
           'consolidation': {'consolidatedSchema': 'MY_SCHEMA',
                            'consolidatedTableName': 'MY_TABLE',
                            'consolidationKeys': ['ID'],
                            'viewSchema': None},
            'fields': [ {'fieldName': 'id',
                        'fields': [],
                        'mapping': {'columnName': 'ID',
                        'columnType': {'nonNull': True,
                                        'precision': 38,
                                        'scale': 0,
                                        'type': 'NUMERIC'},
                                        'isDiscarded': False,
                                        'machineGenerated': False,
                                        'subFields': None}}
                        ]
           'inputObjects': {'12345-asdfg': ['98765-zxcvb']},
           'mapping': {'isDiscarded': False,
                       'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
                       'outputId': 'a1s2d3-f4g5h6',
                       'readOnly': False,
                       'schema': 'MY_SCHEMA',
                       'tableName': 'MY_TABLE_LOG'},
            'mappingMode': 'AUTO_MAP',
            'name': 'MY_SCHEMA.MY_TABLE',
            'origInputLabel': 'production_database',
            'schemaUrls': ['schema?id=12345-asdfg&schema_object=my_table',
                        'schema?id=d=12345-asdfg&sschema_object=deleted_rows'],
            'state': 'MAPPED',
            'usingDefaultMappingMode': False}

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        new_fields = [f for f in mapping['fields']
                      if f['fieldName'] != field_name]
        new_mapping['fields'] = new_fields

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=False,
            show_removed=True,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message=f'REMOVED FILE {field_name} FROM {self.event_name}')

        return new_mapping

    def change_field_mapping_settings(
            self, field_name, new_data_type, truncate=False, non_null=False):
        """ Updates a single filed in a mapping

        :param field_name: The field name to alter
        :param new_data_type: The new datatype for the mapping such as VARCHAR(1024) or INT
        :param truncate: Set to True if the event should be truncated when it's longer than the specific mapping length. Redshift's max VARCHAR is 65535 and Snowflake's max VARCHAR is 16777216.
        :param non_null: Set to true if the field is needs to be not null
        :return: The altered mapping

        The current mapping field settings

        .. code-block:: python

           {'fieldName': 'name',
                        'fields': [],
                        'mapping': {'columnName': 'NAME',
                        'columnType': {'INT':
                                        'nonNull': False,
                                        'truncate': False,
                                        'type': 'VARCHAR'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}

        The new mapping field settings to apply

        .. code-block:: python

           {'fieldName': 'name',
                        'fields': [],
                        'mapping': {'columnName': 'NAME',
                        'columnType': {'length': 1024,
                                        'nonNull': False,
                                        'truncate': True,
                                        'type': 'VARCHAR'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        if 'VARCHAR' in new_data_type.upper():
            new_length = re.search(r'(\d+)', new_data_type).group(0)
            new_data_type = 'VARCHAR'
        else:
            new_length = None

        # Find the field to change and change it
        for field in new_mapping['fields']:
            if field['fieldName'] == field_name:
                field['mapping']['columnType']['length'] = new_length
                field['mapping']['columnType']['type'] = new_data_type
                field['mapping']['columnType']['truncate'] = truncate
                field['mapping']['columnType']['nonNull'] = non_null

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="NEW MAPPING SET")

        return new_mapping

    def change_field_varchar_length(self, field_name, new_length):
        """ Updates only the length of a varchar for a field.

        :param field_name: The field name to alter
        :param new_length: The new length for the varchar
        :return: None

        .. code-block:: python

           {'fieldName': 'name',
                        'fields': [],
                        'mapping': {'columnName': 'NAME',
                        'columnType': {'length': 1024,
                                        'nonNull': False,
                                        'truncate': False,
                                        'type': 'VARCHAR'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}

        .. code-block:: python

           {'fieldName': 'name',
                        'fields': [],
                        'mapping': {'columnName': 'NAME',
                        'columnType': {'length': 16777216,
                                        'nonNull': False,
                                        'truncate': True,
                                        'type': 'VARCHAR'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        for field in new_mapping['fields']:
            if field['fieldName'] == field_name:
                field['mapping']['columnType']['length'] = new_length

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="CHANGED FILED")

        return new_mapping

    def change_field_null_constraint(self, field_name, nonnull=False):
        """ Removes the NULL constraint from a column

        :param field_name: The column for which to remove the constraint
        :param nonnull: The new null setting for the field
        :return: The altered mapping

        The field mapping details with the current nonnull setting

         .. code-block:: python

             {'fieldName': 'id',
             'fields': [],
             'mapping': {'columnName': 'ID',
                        'columnType': {'nonNull': True,
                                        'precision': 38,
                                        'scale': 0,
                                        'type':'NUMERIC'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}


        The field mapping details with the new nonnull setting

         .. code-block:: python

             {'fieldName': 'id',
             'fields': [],
             'mapping': {'columnName': 'ID',
                        'columnType': {'nonNull': False,
                                        'precision': 38,
                                        'scale': 0,
                                        'type':'NUMERIC'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        for f in new_mapping['fields']:
            if f['fieldName'] == field_name:
                f['mapping']['columnType']['nonNull'] = nonnull

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="ADJUSTED NULL CONSTRAINT")

        return new_mapping

    def check_if_consolidation_uses_log(self):
        """ Checks if the event uses a log table

        :param event_name: The name of the event for which to check
        :return: None

        Prints the event name, consolidation key list, the table data is inserted to, and True/False if the insert contains _LOG in the name
        event_name ['consolidation_key_list'] mapping_table_name True/False
        """

        mapping = self.get_mapping_for_event()
        uses_log = '_LOG' in mapping['mapping']['tableName'].upper()
        print(
            self.event_name,
            mapping['consolidation']['consolidationKeys'],
            mapping['mapping']['tableName'],
            uses_log)

        return uses_log

    def copy_mapping(self, new_event):
        """ Copies mapping from the event the class is instantiated with to a new event. Only the name changes

        :param new_event: The name of the event to copy to
        :param print_mapping: Prints the changes if specified
        :param apply_changes: Applies the changes if specified
        :return: The altered mapping
        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        new_mapping['name'] = new_event

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="MAPPING COPIED")

        return new_mapping

    def set_mapping_from_existing_mapping(
            self, new_event_name, new_schema, new_table, new_input_label):
        """ Takes the mapping from the event the class is instantiated with event and uses it to set a new mapping

        :param new_event_name: The event for which to set the new mapping
        :param new_schema: The schema for the new mapping
        :param new_table: The table for the new mapping
        :param new_input_label: The input label of the new mapping
        :return: The altered mapping
        """
        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        new_mapping['consolidation']['consolidatedSchema'] = new_schema
        new_mapping['consolidation']['consolidatedTableName'] = new_table

        new_mapping['mapping']['outputId'] = None
        new_mapping['mapping']['outputHint'] = None
        new_mapping['mapping']['schema'] = new_schema
        new_mapping['mapping']['tableName'] = new_table + '_LOG'

        new_mapping['name'] = new_event_name
        new_mapping['origInputLabel'] = new_input_label
        new_mapping['inputObjects'] = None
        new_mapping['schemaUrls'] = []

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=True,
            show_removed=False,
            show_added=False)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="MAPPING COPIED")

        return new_mapping

    def add_field_to_mapping(
            self, field_name, column_name, data_type, precision=38, scale=0,
            length=16777216, truncate=False, non_null=False):
        """ Adds a single field without sub-fields to the mapping. Works for Snowflake data types only.

        :param field_name: The field name to add
        :param column_name: The name of the column in the target DB
        :param data_type: The data type for the field out of NUMBER, VARIANT, BOOLEAN, VARCHAR, FLOAT, TIMESTAMP
        :param precision: The precision for numeric data
        :param scale: The scale for numeric data
        :param length: The length for a varchar
        :param truncate: Set to True if the event should be truncated when it's longer than the specific mapping length.  Redshift's max VARCHAR is 65535
        :param non_null: Set to true if the field is needs to be not null
        :return: The altered mapping

        Prints the entire existing mapping before the field is added

        .. code-block:: python

          {'autoMappingError': None,
           'consolidation': {'consolidatedSchema': 'MY_SCHEMA',
                            'consolidatedTableName': 'MY_TABLE',
                            'consolidationKeys': ['ID'],
                            'viewSchema': None},
            'fields': [ {'fieldName': 'id',
                        'fields': [],
                        'mapping': {'columnName': 'ID',
                        'columnType': {'nonNull': True,
                                        'precision': 38,
                                        'scale': 0,
                                        'type': 'NUMERIC'},
                                        'isDiscarded': False,
                                        'machineGenerated': False,
                                        'subFields': None}}
                        ]
           'inputObjects': {'12345-asdfg': ['98765-zxcvb']},
           'mapping': {'isDiscarded': False,
                       'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
                       'outputId': 'a1s2d3-f4g5h6',
                       'readOnly': False,
                       'schema': 'MY_SCHEMA',
                       'tableName': 'MY_TABLE_LOG'},
            'mappingMode': 'AUTO_MAP',
            'name': 'MY_SCHEMA.MY_TABLE',
            'origInputLabel': 'production_database',
            'schemaUrls': ['schema?id=12345-asdfg&schema_object=my_table',
                        'schema?id=d=12345-asdfg&schema_object=deleted_rows'],
            'state': 'MAPPED',
            'usingDefaultMappingMode': False}

        Then prints the entire new mapping to set after the field is added. This example adds the NAME field as a VARCHAR

        .. code-block:: python

            {'autoMappingError': None,
           'consolidation': {'consolidatedSchema': 'MY_SCHEMA',
                            'consolidatedTableName': 'MY_TABLE',
                            'consolidationKeys': ['ID'],
                            'viewSchema': None},
            'fields': [ {'fieldName': 'id',
                        'fields': [],
                        'mapping': {'columnName': 'ID',
                        'columnType': {'nonNull': True,
                                        'precision': 38,
                                        'scale': 0,
                                        'type': 'NUMERIC'},
                                        'isDiscarded': False,
                                        'machineGenerated': False,
                                        'subFields': None}},
                        {'fieldName': 'name',
                        'fields': [],
                        'mapping': {'columnName': 'NAME',
                        'columnType': {'length': 16777216,
                                        'nonNull': False,
                                        'truncate': False,
                                        'type': 'VARCHAR'},
                        'isDiscarded': False,
                        'machineGenerated': False,
                        'subFields': None}}
                        ]
           'inputObjects': {'12345-asdfg': ['98765-zxcvb']},
           'mapping': {'isDiscarded': False,
                       'outputHint': '{"table":"my_table","schema":"MY_SCHEMA"}',
                       'outputId': 'a1s2d3-f4g5h6',
                       'readOnly': False,
                       'schema': 'MY_SCHEMA',
                       'tableName': 'MY_TABLE_LOG'},
            'mappingMode': 'AUTO_MAP',
            'name': 'MY_SCHEMA.MY_TABLE',
            'origInputLabel': 'production_database',
            'schemaUrls': ['schema?id=12345-asdfg&schema_object=my_table',
                        'schema?id=d=12345-asdfg&sschema_object=deleted_rows'],
            'state': 'MAPPED',
            'usingDefaultMappingMode': False}

        """

        # Get the mapping and make a deep copy
        mapping = self.get_mapping_for_event()
        new_mapping = copy.deepcopy(mapping)

        # Adjust the copied mapping
        for field in new_mapping['fields']:
            if field_name == field['fieldName']:
                raise Exception('FIELD ALREADY IN MAPPING %s' % field_name)

        if data_type == 'NUMBER':
            new_field = {'fieldName': field_name,
                         'fields': [],
                         'mapping': {'columnName': column_name,
                                     'columnType': {'nonNull': non_null,
                                                    'precision': precision,
                                                    'scale': scale,
                                                    'type': data_type},
                                     'isDiscarded': False,
                                     'machineGenerated': False,
                                     'subFields': None}}

        elif data_type == 'VARCHAR':

            if length > 16777216:
                raise Exception(
                    'Max varchar length is 16777216. You are not allowed to set a varchar of %s' %
                    str(length))

            new_field = {'fieldName': field_name,
                         'fields': [],
                         'mapping': {'columnName': column_name,
                                     'columnType': {'length': length,
                                                    'nonNull': non_null,
                                                    'truncate': truncate,
                                                    'type': data_type},
                                     'isDiscarded': False,
                                     'machineGenerated': False,
                                     'subFields': None}}

        elif data_type == 'BOOLEAN':
            new_field = {'fieldName': field_name,
                         'fields': [],
                         'mapping': {'columnName': column_name,
                                     'columnType': {'nonNull': non_null,
                                                    'type': data_type},
                                     'isDiscarded': False,
                                     'machineGenerated': False,
                                     'subFields': None}}

        elif data_type == 'VARIANT':
            new_field = {'fieldName': field_name,
                         'fields': [],
                         'mapping': {'columnName': column_name,
                                     'columnType': {'nonNull': non_null,
                                                    'type': data_type},
                                     'isDiscarded': False,
                                     'machineGenerated': False,
                                     'subFields': None}}

        elif data_type == 'FLOAT':
            new_field = {'fieldName': field_name,
                         'fields': [],
                         'mapping': {'columnName': column_name,
                                     'columnType': {'nonNull': non_null,
                                                    'type': data_type},
                                     'isDiscarded': False,
                                     'machineGenerated': False,
                                     'subFields': None}}

        elif data_type == 'TIMESTAMP':
            new_field = {'fieldName': field_name,
                         'fields': [],
                         'mapping': {'columnName': column_name,
                                     'columnType': {'nonNull': non_null,
                                                    'type': data_type},
                                     'isDiscarded': False,
                                     'machineGenerated': False,
                                     'subFields': None}}

        else:
            raise Exception(
                'Only SNOWFlAKE data types of NUMBER, VARIANT, BOOLEAN, VARCHAR, FLOAT, TIMESTAMP are allowed')

        new_mapping['fields'].append(new_field)

        # Print and apply the mapping changes
        self._preview_mapping_changes(
            mapping=mapping,
            new_mapping=new_mapping,
            show_matching=False,
            show_changed=False,
            show_removed=False,
            show_added=True)
        self._apply_mapping_changes(
            mapping=new_mapping,
            print_message="NEW MAPPING SET")

        return new_mapping

    def _preview_mapping_changes(
            self, mapping, new_mapping, show_matching, show_changed,
            show_removed, show_added):
        """ Takes an original mapping and altered mapping and prints various views on the changes

        :param mapping: A dictionary representing the current state of the mapping
        :param new_mapping: An altered dictionary representing the changed state of the mapping
        :param show_matching: Show the key value pairs that match between the two mappings
        :param show_changed: Show the key value pairs that have been changed between the two mappings
        :param show_removed: Show the key value pairs that were removed from the current state of the mapping
        :param show_added: Show the key value pairs that were added to the changed state of the mapping
        :return: None
        """

        # Instantiate the Dictionary Difference Class
        DD = DictionaryDifferences(
            old_dictionary=mapping,
            new_dictionary=new_mapping,
            pprint_indent=self.pprint_indent,
            pprint_width=self.pprint_width,
            pprint_depth=self.pprint_depth)

        # Print the full before and after dictionaries
        if self.preview_full:
            DD.show_dictionary_all()

        # Print only the differences between the dictionaries
        if self.preview_changes:
            DD.show_dictionary_differences(
                show_matching=show_matching,
                show_changed=show_changed,
                show_removed=show_removed,
                show_added=show_added)

        return None

    def _apply_mapping_changes(self, mapping, print_message):
        """ Apply the mapping changes in the Alooma API

        :param mapping: The new mapping to set
        :param print_message: A message to print after a successful change
        :return: None
        """

        if self.apply_changes:
            try:
                self.api.set_mapping(mapping, self.event_name)
                print(print_message, self.event_name)
            except Exception as e:
                print(e)

        return None

    # Needs tests
    def remove_unmapped_fields_and_clear_table_stats(self):
        """ Remove unmapped fields from the a mapping and clear the UI stats.
        The mapper can become very slow over time and clearing the table stats periodically will help speed it back up.

        :return: none
        """

        if self.preview_full:
            print(
                "CLEAR UNMAPPED FIELDS AND RESET STATS IN UI FOR",
                self.event_name)

        if self.apply_changes:
            url = self.api.rest_url + "event-types/%s/clear-stats" % self.event_name
            response = self.api._Client__send_request(
                requests.delete, url, timeout=300)

            if response.status_code == 204:
                print(
                    "REMOVED UNMAPPED FIELDS AND RESET TABLE STATS IN UI",
                    self.event_name)
            else:
                raise Exception(
                    'Unable to clear stats for ' + self.event_name +
                    ' response code is ' + str(response.status_code))

            return response.status_code



import pprint
import json
import requests
import copy
import pandas as pd
from tabulate import tabulate


from .dictionary_difference import DictionaryDifferences


class Inputs:

    def __init__(self, api, preview_full=True, preview_changes=True, apply_changes=False, pprint_indent=2, pprint_width=250, pprint_depth=5):
        """View, created and edit Alooma inputs. Inputs are the source of data to bring into Alooma.
        The class is initiated with the following variables:

        :param api: The Alooma API client authentication
        :param preview_full: Prints the input and changed input if True
        :param preview_changes: Prints the changes in the input by category if True
        :param apply_changes: Executes the mapping changes if True
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        :param pprint_depth: The depth value to pprint dictionaries
        """

        self.api = api
        self.preview_full = preview_full
        self.preview_changes = preview_changes
        self.apply_changes = apply_changes
        self.pprint_indent = pprint_indent
        self.pprint_width = pprint_width
        self.pprint_depth = pprint_depth

    @staticmethod
    def print_sorted_list(lst):
        """
        :param lst: list Takes a list and prints a sorted list
        """
        print('\n'.join(sorted(lst)))

    def get_all_inputs(self):
        """ Gets the information for all inputs from Alooma

        :return: A list of dictionaries with information for each input
        """
        return self.api.get_inputs()

    def get_input(self, input_name):
        """ Gets the input information for a specific input

        :param input_name: The name of the input to retrieve
        :return: A dictionary with the input
        """
        input_list = self.get_all_inputs()
        single_input = [i for i in input_list if input_name == i['name']][0]

        return single_input

    def view_inputs(self, single_input=None, print_format='table'):
        """ Prints a data from with the event info

        :param single_input: string The name of a specific input to filter the results
        :param print_format: string 'table' to print event info as tables or 'json' to print as dictionaries
        :return: The list of input dictionaries
        """

        inputs = self.get_all_inputs()

        if single_input:
            inputs = [i for i in inputs if i["name"] == single_input]

        if print_format == 'table':
            df = pd.DataFrame(inputs)
            df = df.reindex(['category', 'name', 'type', 'deleted', 'paused', 'stats', 'state', 'id', 'configuration'], axis=1)
            df.sort_values(by=['name'], ascending=True, inplace=True)
            print(tabulate(df, headers='keys', tablefmt='psql', showindex=False))
            return df
        elif print_format == 'json':
            for i in inputs:
                pprint.pprint(i, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)
            if single_input:
                return inputs[0]
            else:
                return inputs

    # Needs tests
    def delete_input(self, input_name):
        """ Delete an input. This can not be undone!

        :param input_name: The name of the input to delete
        :return: None
        """
        input = self.get_input(input_name)

        if self.preview_full:
            print("INPUT TO DELETE")
            pprint.pprint(input, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        if self.apply_changes:
            self.api.remove_input(input['id'])
            print("INPUT DELETED", input_name)

        return None

    # Needs tests
    def pause_input(self, input_name):
        """ Pause the input from pulling data

        :param input_name: The name of the input to pause
        :return: None
        """

        input = self.get_input(input_name)

        params = {"pause": 'true'}
        url = self.api.rest_url + "inputs/{input_id}/pause".format(input['id'])

        res = self.api._Client__send_request(requests.put, url, params=params)
        print("Paused Input:", input_name, res)

        return None

    # Needs tests
    def resume_input(self, input_name):
        """ Resume the input so it will pull date

        :param input_name: The name of the input to resume pulling data
        :return: None
        """

        input = self.get_input(input_name)
        params = {"pause": 'false'}
        url = self.api.rest_url + "inputs/{input_id}/pause".format(input['id'])
        res = self.api._Client__send_request(requests.put, url, params=params)
        print("Resumed Input:", input_name, res)

        return None

    def list_inputs(self, add_quotes_commas=False):
        """ Prints and/or returns a list of input names

        :param print_lst: boolean True to print the list to the console
        :param add_quotes_commas: boolean True to print the list to the console with single quotes around strings and commas between each name
        :return: list of the names of inputs
        """

        inputs = self.get_all_inputs()
        input_name_list = [i["name"] for i in inputs]

        if self.preview_full:
            if not add_quotes_commas:
                self.print_sorted_list(input_name_list)
            else:
                input_name_list = ["'" + i + "'," for i in inputs]
                self.print_sorted_list(input_name_list)

        return input_name_list

    def list_tables(self, input_name):
        """ Prints and/or returns the list of tables in an input

        :param single_input: string The input for which to print or return the tables
        :return: The list if tables in an input
        """

        single_input = self.get_input(input_name=input_name)

        if single_input['type'] != 'ODBC':
            raise Exception('The input ' + single_input['name'] + ' is of type ' + single_input['type'] + ' and not of type ODBC')
        else:
            tables = single_input['configuration']['tables']

        tables = json.loads(tables)
        table_lst = list(tables.keys())

        if self.preview_full:
            self.print_sorted_list(table_lst)

        return table_lst

    def create_input_database(self, source_credentials, new_input_name, existing_input, tables_dict=None, auto_map=True, input_default_schema=None, replication_type='incremental_load', batch_size=100000):
        """ Create a new input with a database as the source

        :param source_credentials: The database credentials and configuration for the source
        :param new_input_name: The name of the new input
        :param existing_input: The name of the existing input from which to copy the configuration
        :param tables_dict: A dictionary containing the tables to replicate and the update indicator such as {"table_name": "xmin::text::bigint"}
        :param auto_map: True if the input should be auto-mapped
        :param input_default_schema: The default schema for the data in the target database
        :param replication_type: The type of replication to apply
        :param batch_size: The number of rows to pull in each batch
        :return: None

        Example database source credentials:

        .. code-block:: python

            {'server': 'server_name',
             'schema': 'schema_name',
             'port': 'port',
             'database': 'database',
             'db_type': 'psql',
             'user': 'username',
             'password': 'password'}

        """

        single_input = self.get_input(input_name=existing_input)

        if single_input['type'] != 'ODBC':
            raise Exception('The input ' + single_input['name'] + ' is of type ' + single_input['type'] + ' and not of type ODBC')

        new_input_config = {}

        # Source Database Information
        new_input_config['server'] = source_credentials['server']
        new_input_config['schema'] = 'public'
        new_input_config['port'] = source_credentials['port']
        new_input_config['database'] = source_credentials['database']
        new_input_config['db_type'] = source_credentials['db_type']
        new_input_config['user'] = source_credentials['user']
        new_input_config['password'] = source_credentials['password']

        # Target database and configuration information
        new_input_config['auto_map'] = auto_map
        new_input_config['batch_size'] = batch_size
        new_input_config['db_type'] = input_default_schema
        new_input_config['replication_type'] = replication_type
        new_input_config['input_default_schema'] = input_default_schema
        new_input_config['tables'] = json.dumps(tables_dict)

        if self.preview_full:
            print("EXISTING INPUT SAMPLE")
            pprint.pprint(single_input, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

            print("NEW INPUT CONFIG")
            pprint.pprint(new_input_config, indent=self.pprint_indent, width=self.pprint_width, depth=self.pprint_depth)

        if self.apply_changes:
            self.api.create_input({'name': new_input_name,
                                   'type': 'ODBC',
                                   'configuration': new_input_config})
            print("NEW INPUT CREATED")

        return new_input_config

    def edit_input_configuration(self, input_name=None, field_to_edit=None, new_field_value=None):
        """ Edit a single field in an input configuration at a time

        :param input_name: The name of the input to edit
        :param field_to_edit: The name of the field to edit
        :param new_field_value: The new value for the field
        :return: None
        """

        # Get the mapping and make a deep copy
        input = self.get_input(input_name=input_name)
        new_input = copy.deepcopy(input)

        if field_to_edit not in new_input['configuration']:
            raise Exception('The field ' + field_to_edit + ' is not in the input ' + input_name)

        # Adjust the copied input
        new_input['configuration'][field_to_edit] = new_field_value

        # Print and apply the input changes
        self.preview_input_changes(input=input, new_input=new_input, show_matching=False, show_changed=True, show_removed=False, show_added=False)
        self.apply_input_changes(input=new_input, print_message="CHANGED THE INPUT CONFIGURATION")

        return new_input

    def add_table_to_input(self, input_name, new_table_dict={}):
        """ Add tables to an existing input

        :param input_name: The name of the input to add tables to
        :param new_table_dict: A dictionary with the tables to add and their update field such as  {"table_name": "xmin::text::bigint"}
        :return: None
        """

        # Get the existing input and make a deep copy
        single_input = self.get_input(input_name=input_name)
        new_input = copy.deepcopy(single_input)

        # Check that the input type is ODBC
        if single_input['type'] != 'ODBC':
            raise Exception('The input ' + single_input['name'] + ' is of type ' + single_input['type'] + ' and not of type ODBC')

        # Load the existing string tables to a dictionary
        table_dict = json.loads(single_input['configuration']['tables'])

        for table in new_table_dict:
            if table in table_dict:
                raise Exception('The table ' + table + ' is already in the input ' + input_name)
            else:
                table_dict[table] = new_table_dict[table]

        new_input['configuration']['tables'] = json.dumps(table_dict)

        # Print and apply the input changes
        self.preview_input_changes(input=single_input, new_input=new_input, show_matching=False, show_changed=True, show_removed=False, show_added=False)
        self.apply_input_changes(input=new_input, print_message="ADDED TABLES TO INPUT")

        return new_input

    def change_auto_mapping_mode(self, input_name=None, new_mapping_mode=False):
        """ Edit a single field in an input configuration at a time

        :param input_name: The name of the input to alter
        :param new_mapping_mode: The new automapping mapping mode to set of True or False
        :return: None
        """

        # Get the mapping and make a deep copy
        input = self.get_input(input_name=input_name)
        new_input = copy.deepcopy(input)

        # Adjust the copied input
        new_input['configuration']['auto_map'] = new_mapping_mode

        # Print and apply the input changes
        self.preview_input_changes(input=input, new_input=new_input, show_matching=False, show_changed=True, show_removed=False, show_added=False)
        self.apply_input_changes(input=new_input, print_message="CHANGED AUTO MAPPING MODE")

        return new_input

    def add_template_to_parameter_configuration(self, input_name, add_to_parameter, template):
        """ Add a template to single parameter or a list of parameters in an input configuration

        :param input_name: The input to add template to
        :param add_to_parameter: a single parameter or a list of parameters that you want to add template to
        :param template: e.g. '%Y-%m-%d'
        :return: None
        """

        # Get the mapping and make a deep copy
        input = self.get_input(input_name=input_name)
        new_input = copy.deepcopy(input)

        if 'parameters' not in input['configuration']:
            raise Exception('The input ' + input_name + ' does not have any parameters')

        # Adjust the copied input
        for param in input['configuration']['parameters']:
            if param['parameter'] in add_to_parameter:
                param['template'] = template

        # Print and apply the input changes
        self.preview_input_changes(input=input, new_input=new_input, show_matching=False, show_changed=True, show_removed=False, show_added=False)
        self.apply_input_changes(input=new_input, print_message="ADDED TEMPLATE TO CONFIGURATION PARAMETER")

        return new_input

    def edit_parameter_configuration(self, input_name, parameter_to_edit, value_to_set, new_value):
        """ edit API parameter value, e.g. changing days past value from 100 to 10

        :param input_name: The input to make the change for
        :param parameter_to_edit: a single parameter that you want to edit
        :param value_to_set: name of the field you want to edit
        :param new_value: new value you want to set
        :return: None
        """

        # Get the mapping and make a deep copy
        input = self.get_input(input_name=input_name)
        new_input = copy.deepcopy(input)

        if 'parameters' not in input['configuration']:
            raise Exception('The input ' + input_name + ' does not have any parameters')

        # Adjust the copied input
        for param in new_input['configuration']['parameters']:
            if param['parameter'] == parameter_to_edit:
                param[value_to_set] = new_value

        # Print and apply the input changes
        self.preview_input_changes(input=input, new_input=new_input, show_matching=False, show_changed=True, show_removed=False, show_added=False)
        self.apply_input_changes(input=new_input, print_message="INPUT PARAMETER CONFIGURATION CHANGED")

        return new_input

    def preview_input_changes(self, input, new_input, show_matching, show_changed, show_removed, show_added):
        """ Takes an original mapping and altered mapping and prints various views on the changes

        :param input: A dictionary representing the current state of the mapping
        :param new_input: An altered dictionary representing the changed state of the mapping
        :param show_matching: Show the key value pairs that match between the two mappings
        :param show_changed: Show the key value pairs that have been changed between the two mappings
        :param show_removed: Show the key value pairs that were removed from the current state of the mapping
        :param show_added: Show the key value paids that were added to the changed state of the mapping
        :return: None
        """

        # Instantiate the Dictionary Difference Class
        DD = DictionaryDifferences(old_dictionary=input, new_dictionary=new_input, pprint_indent=self.pprint_indent, pprint_width=self.pprint_width, pprint_depth=self.pprint_depth)

        # Print the full before and after dictionaries
        if self.preview_full:
            DD.show_dictionary_all()

        # Print only the differences between the dictionaries
        if self.preview_changes:
            DD.show_dictionary_differences(show_matching=show_matching, show_changed=show_changed, show_removed=show_removed, show_added=show_added)

        return None

    def apply_input_changes(self, input, print_message):
        """ Apply the input changes in the Alooma API

        :param input: The new input to set
        :param print_message: A message to print after a successful change
        :return: None
        """

        if self.apply_changes:
            try:
                self.api.edit_input(input)
                print(print_message, self.input_name)
            except Exception as e:
                print(e)

        return None



class ColumnDDL:
    """
    Prints DDL for direct DB changes of data type, dropping columns, or adding columns
    """

    def __init__(self, tuple_or_tuple_list, change_type='data_type', has_log=False, case='UPPER'):
        """ Prints DDL for direct DB changes of data_type, drop, or add

        :param tuple_or_tuple_list: A single tuple or list of tuples with information about the changes to make
        :param change_type: The type of DDL to create out of data_type, add, rename, combine or drop
        :param log_table: True if the event has a log table
        :param case: UPPER or LOWER case for the DDL statements

        For a change_type of data_type or adding a column the tuple_or_tuple_list should contain the schema.table, the column name, and the column data type.

        .. code-block:: python

            ('schema_name.table_name', 'column_one', 'INT')
            or
            [('schema_name.table_name', 'column_one', 'INT'),
            ('schema_name.table_name', 'column_two', 'VARCHAR(16777216)')]

        For a change_type of dropping a column a column the tuple_or_tuple_list should contain the schema.table and the column name

        .. code-block:: python

            ('schema_name.table_name', 'column_one')
            or
            [('schema_name.table_name', 'column_one'),
            ('schema_name.table_name', 'column_two')]

        For a change_type of renaming a column a column the tuple_or_tuple_list should contain the schema.table, the current column name, and the new column name

        .. code-block:: python

            ('schema_name.table_name', 'my_bad_column', 'my_good_column', 'INT')
            or
            [('schema_name.table_name', 'my_bad_column_one', 'my_good_column_one', 'INT'),
            ('schema_name.table_name', 'my_bad_column_two', 'my_good_column_two', 'VARCHAR(16777216)')]

        For a change_type of combining columns the tuple_or_tuple_list should contain the schema.table, the current column name, the new column name, and the column data type.

        .. code-block:: python

            ('schema_name.table_name', 'my_bad_column', 'my_good_column', 'INT')
            or
            [('schema_name.table_name', 'my_bad_column_name_one', 'my_good_column_name_one', 'INTEGER'),
            ('schema_name.table_name', 'my_bad_column_name_two', 'my_good_column_name_two', 'VARCHAR(16777216)')]

        """

        self.tuple_or_tuple_list = tuple_or_tuple_list
        self.change_type = change_type
        self.has_log = has_log
        self.case = case

    def convert_tuple_to_list(self):
        """ Converts single tuples into a list of tuples so list iteration on single event changes works

        :param tuple_or_tuple_list: A tuple or a tuple list
        :return: A list of tuple
        """

        # Check if the input is a single tuple and convert to a list if needed
        if type(self.tuple_or_tuple_list) is tuple:
            column_list = []
            column_list.append(self.tuple_or_tuple_list)
        else:
            column_list = self.tuple_or_tuple_list

        return column_list

    def create_ddl_statements(self):
        """ Creates DDL statements, prints the statements, and returns the statements

        :param event_column_type_tuple: The tuple from the function specified
        :param change_type: The type of column change statements to print. Options are column, drop, combine_drop
        :param log_table: True if the columns to alter have log tables
        :return: A string with all the queries generated

        """

        # The query list returns all the queries generate and is used for testing
        query_strings = ''

        # Check if the input is a single tuple and convert to a list if needed
        column_list = self.convert_tuple_to_list()

        # Parse the tuples into pieces and print the DDL statements
        for column in column_list:
            fully_qualified_table_name_db = column[0]
            split_fully_qualified_table_name_db = fully_qualified_table_name_db.split('.')
            schema_name = split_fully_qualified_table_name_db[0]
            table_name = split_fully_qualified_table_name_db[1]

            column_name = column[1]

            if len(column) >= 3 and self.change_type in ['add', 'data_type']:
                column_data_type = column[2]

                if 'varchar' in column_data_type:
                    column_data_type_no_char_count = 'varchar'
                else:
                    column_data_type_no_char_count = column_data_type

            if len(column) == 3 and self.change_type == 'rename':
                new_column_name = column[2]

            if len(column) >= 4 and self.change_type == 'combine':
                new_column_name = column[2]
                column_data_type = column[3]

                if 'varchar' in column_data_type:
                    column_data_type_no_char_count = 'varchar'
                else:
                    column_data_type_no_char_count = column_data_type

            if self.change_type == 'data_type':
                query = self.change_column_data_type(schema_name=schema_name, table_name=table_name, column_name=column_name, new_column_type=column_data_type, new_column_type_no_count=column_data_type_no_char_count)
            if self.change_type == 'drop':
                query = self.drop_column(schema_name=schema_name, table_name=table_name, column_to_drop=column_name)
            if self.change_type == 'add':
                query = self.add_column(schema_name=schema_name, table_name=table_name, column_to_add=column_name, new_column_type=column_data_type)
            if self.change_type == 'rename':
                query = self.rename_column(schema_name=schema_name, table_name=table_name, existing_column_name=column_name, new_column_name=new_column_name)
            if self.change_type == 'combine':
                query = self.combine_columns(schema_name=schema_name, table_name=table_name, column_to_drop=column_name, column_to_keep=new_column_name, column_to_keep_type=column_data_type_no_char_count)

            if self.case == 'UPPER':
                query = query.upper()
            else:
                query = query.lower()

            print(query)

            query_strings = query_strings + query

        return query_strings

    def drop_column(self, schema_name, table_name, column_to_drop):
        """ Prints drop column statements

        :param schema_name: The name of the schema
        :param table_name: The name of the table
        :param column_to_drop: The name of the column to drop
        :return: the query

        For a change_type of dropping a column a column the tuple_or_tuple_list should contain the schema.table and the column name

        .. code-block:: python

            ('schema_name.table_name', 'column_one')
            or
            [('schema_name.table_name', 'column_one'),
            ('schema_name.table_name', 'column_two')]

        """

        if self.has_log:
            query = f"""BEGIN TRANSACTION;
                         ALTER TABLE "{schema_name}"."{table_name}"
                         DROP COLUMN "{column_to_drop}" CASCADE;

                         ALTER TABLE "{schema_name}"."{table_name}_LOG"
                         DROP COLUMN "{column_to_drop}" CASCADE ;

                        COMMIT;"""

        else:
            query = f"""BEGIN TRANSACTION;
                         ALTER TABLE "{schema_name}"."{table_name}"
                         DROP COLUMN "{column_to_drop}" CASCADE;

                        COMMIT;"""

        return query

    def add_column(self, schema_name, table_name, column_to_add, new_column_type):
        """ Prints add column statements

        :param schema_name: The name of the schema
        :param table_name: The name of the table
        :param column_to_add: The name of the column to add
        :return: The query

        For a change_type of data_type or adding a column the tuple_or_tuple_list should contain the schema.table, the column name, and the column data type.

        .. code-block:: python

            ('schema_name.table_name', 'column_one', 'INT')
            or
            [('schema_name.table_name', 'column_one', 'INT'),
            ('schema_name.table_name', 'column_two', 'VARCHAR(16777216)')]


        """

        if self.has_log:
            query = f"""ALTER TABLE "{schema_name}"."{table_name}"
                        ADD COLUMN "{column_to_add}" {new_column_type};
                              
                        ALTER TABLE "{schema_name}"."{table_name}_LOG"
                        ADD COLUMN "{column_to_add}" {new_column_type};"""

        else:
            query = f"""ALTER TABLE "{schema_name}"."{table_name}"
                        ADD COLUMN "{column_to_add}" {new_column_type};"""

        return query

    def change_column_data_type(self, schema_name, table_name, column_name, new_column_type, new_column_type_no_count):
        """ Prints the DB DDL statement for a single column

        :param schema_name: The name of the schema
        :param table_name: The name of the table
        :param log_table: True if the table has a log table
        :return: The query

        For a change_type of data_type or adding a column the tuple_or_tuple_list should contain the schema.table, the column name, and the column data type.

        .. code-block:: python

            ('schema_name.table_name', 'column_one', 'INT')
            or
            [('schema_name.table_name', 'column_one', 'INT'),
            ('schema_name.table_name', 'column_two', 'VARCHAR(16777216)')]


        """

        if self.has_log:
            query = f"""
                     BEGIN TRANSACTION;
                         ALTER TABLE "{schema_name}"."{table_name}"
                         RENAME COLUMN "{column_name}" TO "{column_name}_TMP";

                         ALTER TABLE "{schema_name}"."{table_name}"
                           ADD COLUMN "{column_name}" {new_column_type};
                           UPDATE "{schema_name}"."{table_name}"
                           SET "{column_name}" = "{column_name}_TMP"::{new_column_type_no_count}
                           WHERE "{column_name}" IS NULL;

                         ALTER TABLE "{schema_name}"."{table_name}"
                           DROP COLUMN "{column_name}_TMP" CASCADE;

                         ALTER TABLE "{schema_name}"."{table_name}_LOG"
                         RENAME COLUMN "{column_name}" TO "{column_name}_TMP";

                          ALTER TABLE "{schema_name}"."{table_name}_LOG"
                           ADD COLUMN "{column_name}" {new_column_type};
                           UPDATE "{schema_name}"."{table_name}_LOG"
                           SET "{column_name}" = "{column_name}_TMP"::{new_column_type_no_count}
                           WHERE "{column_name}" IS NULL;

                         ALTER TABLE "{schema_name}"."{table_name}_LOG"
                           DROP COLUMN "{column_name}_TMP" CASCADE;
                     COMMIT;
                     """

        else:
            query = f"""BEGIN TRANSACTION;
                         ALTER TABLE "{schema_name}"."{table_name}"
                         RENAME COLUMN "{column_name}" TO "{column_name}_TMP";

                         ALTER TABLE "{schema_name}"."{table_name}"
                           ADD COLUMN "{column_name}" {new_column_type};
                           UPDATE "{schema_name}"."{table_name}"
                           SET "{column_name}" = "{column_name}_TMP"::{new_column_type_no_count}
                           WHERE "{column_name}" IS NULL;

                         ALTER TABLE "{schema_name}"."{table_name}"
                           DROP COLUMN "{column_name}_TMP" CASCADE;
                        COMMIT;"""
        return query

    def rename_column(self, table_name, schema_name, existing_column_name, new_column_name):
        """ Prints the DB DDL statement to rename a single column

        :param schema_name: The name of the schema
        :param table_name: The name of the table
        :param existing_column_name: The existing column name of the colum
        :param new_column_name: The new column name for the column
        :return: The query

        For a change_type of renaming a column a column the tuple_or_tuple_list should contain the schema.table, the current column name, and the new column name

        .. code-block:: python

            ('schema_name.table_name', 'my_bad_column', 'my_good_column', 'INT')
            or
            [('schema_name.table_name', 'my_bad_column_one', 'my_good_column_one', 'INT'),
            ('schema_name.table_name', 'my_bad_column_two', 'my_good_column_two', 'VARCHAR(16777216)')]

        """

        if self.has_log:
            query = f"""ALTER TABLE "{schema_name}"."{table_name}"
                         RENAME COLUMN "{existing_column_name}" TO "{new_column_name}";

                         ALTER TABLE "{schema_name}"."{table_name}_LOG"
                         RENAME COLUMN "{existing_column_name}" TO "{new_column_name}";"""

        else:
            query = f"""ALTER TABLE "{schema_name}"."{table_name}"
                         RENAME COLUMN "{existing_column_name}" TO "{new_column_name}";"""
        return query

    def combine_columns(self, schema_name, table_name, column_to_drop, column_to_keep, column_to_keep_type):
        """ Prints statement to combine data from 2 columns into a target column and drop the non-target column

        :param fully_qualified_table_name_db: The schema.table_name of the table to change
        :param column: The tuple for an individual column
        :param log_table: True if the table has a log table
        :return: The query

        For a change_type of combining columns the tuple_or_tuple_list should contain the schema.table, the current column name, the new column name, and the column data type.

        .. code-block:: python

            ('schema_name.table_name', 'my_bad_column', 'my_good_column', 'INT')
            or
            [('schema_name.table_name', 'my_bad_column_name_one', 'my_good_column_name_one', 'INTEGER'),
            ('schema_name.table_name', 'my_bad_column_name_two', 'my_good_column_name_two', 'VARCHAR(16777216)')]

        """

        if self.has_log:
            query = f"""BEGIN TRANSACTION;
                         UPDATE TABLE "{schema_name}"."{table_name}"
                           SET "{column_to_keep}" = "{column_to_drop}"::{column_to_keep_type}
                           WHERE "{column_to_keep}" IS NULL;

                         ALTER TABLE "{schema_name}"."{table_name}"
                         DROP COLUMN "{column_to_drop}" CASCADE;

                          UPDATE TABLE "{schema_name}"."{table_name}_LOG"
                           SET "{column_to_keep}" = "{column_to_drop}"::{column_to_keep_type}
                           WHERE "{column_to_keep}" IS NULL;

                         ALTER TABLE "{schema_name}"."{table_name}_LOG"
                         DROP COLUMN "{column_to_drop}" CASCADE;
                     COMMIT;
                     """

        else:
            query = f"""BEGIN TRANSACTION;
                            UPDATE TABLE "{schema_name}"."{table_name}"
                               SET "{column_to_keep}" = "{column_to_drop}"::{column_to_keep_type}
                               WHERE "{column_to_keep}" IS NULL;
    
                             ALTER TABLE "{schema_name}"."{table_name}"
                             DROP COLUMN "{column_to_drop}" CASCADE;
                        COMMIT;"""
        return query




import os
import pprint
import json
import os.path
import requests
from requests.auth import HTTPBasicAuth
from sqlalchemy.engine.url import URL
from sqlalchemy import create_engine
import pandas as pd
import datetime

from .events import Events
from .inputs import Inputs


class Samples:

    def __init__(self, api, sample_event_directory):
        """ For each event Alooma stores sample dictionaries. This class will help you interact and copy the samples for testing. We run integration testing on the entirety of the sample set before deploying.

        :param api: The Alooma API client authentication
        :param sample_event_directory: The full path and name of the directory in which to store sample events such as /Users/myname/code/hover/alooma-etl/sample_events
        """
        self.api = api
        self.sample_event_directory = sample_event_directory

    def build_sample_file_path_and_file_name(self, sample_file_type='event', name=None):
        """ Constructs the file path and name for writing the samples

        :param sample_file_type: The type of file either 'event' or 'input'
        :param name: The name of the event or input
        :return: a file path with the file name
        """

        file_name = sample_file_type + '_' + name.replace('.', '_').lower() + '.json'
        file_path = os.path.join(self.sample_event_directory, file_name)

        return file_path

    @staticmethod
    def print_samples(sample_list, view_all_or_index='all', pprint_indent=2, pprint_width=200, pprint_depth=5):
        """ Prints all the samples in a list or pretty prints a single sample

        :param sample_list: A list of samples
        :param view_all_or_index: 'all' to print all samples or a number to print sample by index
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        :param pprint_depth: The depth value to pprint dictionaries
        :return: None
        """

        if view_all_or_index == 'all':
            for idx, val in enumerate(sample_list):
                print(idx, val)
        else:
            pprint.pprint(sample_list[int(view_all_or_index)], indent=pprint_indent, width=pprint_width, depth=pprint_depth)

    def get_samples_for_event_from_alooma_api(self, event_name):
        """ Gets samples for a single event from the API

        :param event_name: The event for which to get samples
        :return: A list of samples
        """

        sample_list = self.api.get_samples(event_type=event_name)
        return sample_list

    def view_samples_for_event_from_alooma_api(self, event_name, view_all_or_index='all'):
        """ View samples for an event from the API. Print all or a single event by index.

        :param event_name: The name of the event for which to view samples
        :param view_all_or_index: Specify to view 'all' the samples for an event or a specific event by index number
        """
        samples = self.get_samples_for_event_from_alooma_api(event_name=event_name)
        self.print_samples(samples, view_all_or_index)

        return samples

    @staticmethod
    def get_samples_from_file(file_path):
        """

        :return:
        """
        with open(file_path, 'r') as outfile:
            sample_list = json.loads(outfile.read())

        return sample_list

    def get_samples_from_saved_sample_files(self, sample_file_type='event', name=None):
        """ Gets the samples for an event or input from a saved file

        :param sample_file_type: The name of the file to retrieve samples for. Exclude the .json extension
        :param name: The name of the event or input to get the samples for
        :return: A list with the samples from the file
        """

        if sample_file_type == 'event':
            file_path = self.build_sample_file_path_and_file_name(sample_file_type='event', name=name)
        elif sample_file_type == 'input':
            file_path = self.build_sample_file_path_and_file_name(sample_file_type='input', name=name)

        return self.get_samples_from_file(file_path)

    def view_samples_from_file(self, sample_file_type='event', name=None, view_all_or_index='all', pprint_indent=2, pprint_width=200, pprint_depth=5):
        """ Prints the samples for an event or input from the API

        :param sample_file_type: Whether the samples are for an 'event' or an 'input'
        :param name: The name of the event or input
        :param view_all_or_index: Specify to view 'all' the samples for an event or a specific event by index
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        :param pprint_depth: The depth value to pprint dictionaries
        :return: None
        """

        samples = self.get_samples_from_saved_sample_files(sample_file_type=sample_file_type, name=name)
        self.print_samples(samples, view_all_or_index, pprint_indent=pprint_indent, pprint_width=pprint_width, pprint_depth=pprint_depth)

        return samples

    def save_alooma_samples_to_files(self, event_name=None, input_name=None):
        """ Saves samples to files in the specified directory.
        If no event_name or input_name are specified samples for ALL events will be written to a file

        :param event_name: The event name for samples to save.  If specified only samples for this event are retrieved
        :param input_name: The input name for samples to save. If specified all samples for this input are retrieved
        :return: None
        """

        E = Events(api=self.api)
        I = Inputs(api=self.api)

        # Writes samples to a file for the event
        if event_name is not None:
            self.write_alooma_samples_to_files(event_list=[event_name], input_name=None)

        # Writes samples to a file for each event and a file with all events for the input
        elif input_name is not None:
            event_list = E.list_events(print_lst=False, input_labels=input_name)
            self.write_alooma_samples_to_files(event_list=event_list, input_name=input_name)

        # Writes samples to a file for each event and all events for an input to a file for each input
        else:
            input_list = I.list_inputs(print_lst=False)

            for input_name in input_list:
                event_list = E.list_events(print_lst=False, input_labels=input_name)
                self.write_alooma_samples_to_files(event_list=event_list, input_name=input_name)

        return None

    def write_alooma_samples_to_files(self, event_list, input_name=None):
        """ Writes samples to files for each event. Writes input file with samples for all events in the input when input_name is specified.

        :param event_list: A list of events for which to write samples for
        :param input_name: The name of the input to write the samples for to limit the samples to a specific input
        :return: None
        """

        sample_lst = []

        for event in event_list:
            event_file_path = self.build_sample_file_path_and_file_name(sample_file_type='event', name=event)

            print("GETTING & WRITING SAMPLES FOR:", event)

            samples = self.get_samples_for_event_from_alooma_api(event)

            with open(event_file_path, 'w') as outfile:
                json.dump(samples, outfile)

            sample_lst = sample_lst + samples

        if input_name is not None:
            print("WRITING INPUT SAMPLES FOR:", input_name)
            print("INPUT SAMPLE COUNT:", len(sample_lst))
            input_file_path = self.build_sample_file_path_and_file_name(sample_file_type='input', name=input_name)

            with open(input_file_path, 'w') as outfile:
                json.dump(sample_lst, outfile)

        return None

    @staticmethod
    def get_sample_from_any_api(url, payload, api_key, input_label, print_api_response=True, test_index_number=0, input_type='rest_endpoint'):
        """ Gets a sample to test from a 3rd party API using the requests package and adds the required Alooma _metadata to the event

        :param url: The URL for which to make the get request e.g. https://harvest.greenhouse.io/v1/scheduled_interviews
        :param payload: A JSON dictionary with the parameters to add to the requests e.g. {'updated_after': '2018-08-25T00:00:00Z'}
        :param api_key: The API key with which to make the request
        :param preview_events: If True all the events retrieved will be printed to the console
        :param input_label: The input_label to give these events when deployed in Alooma
        :param test_index_number: The index number of the event to test
        :param input_type: The type of input the data will use when deployed to Alooma
        :return: An event with metadata added that can be tested through the transformation code
        """

        try:
            response = requests.get(url, auth=HTTPBasicAuth(api_key, api_key), params=payload, timeout=3)
            response.raise_for_status()
        except requests.exceptions.HTTPError as errh:
            raise Exception('Http Error: %s' % errh)
        except requests.exceptions.ConnectionError as errc:
            raise Exception('Error Connecting:' % errc)
        except requests.exceptions.Timeout as errt:
            raise Exception('Timeout Error: %s' % errt)
        except requests.exceptions.RequestException as err:
            raise Exception('Timeout Error: %s' % err)

        events = response.json()

        if print_api_response:
            for event in events:
                print(event)

        event = events[test_index_number]
        event['_metadata'] = {'input_label': input_label, 'input_type': input_type}

        return event

    @staticmethod
    def get_sample_from_postgresql_database(db_host, db_port, db_user, db_password, db_name, input_label, table_name, id_field, row_id, print_event=True):

        db_settings = {
            'drivername': 'postgres',
            'host': db_host,
            'port': db_port,
            'username': db_user,
            'password': db_password,
            'database': db_name
        }

        # Adds quotes to the ID field if it's a string
        if isinstance(row_id, str):
            row_id = "\'" + row_id + "\'"

        # Creates the query, gets a row from PostgreSQL,  and adds the Alooma _metadata needed to go through the code
        engine = create_engine(URL(**db_settings))
        sql = "SELECT * FROM {table_name} WHERE {id_field} = {row_id};".format(table_name=table_name, id_field=id_field, row_id=row_id)
        df = pd.read_sql(sql, engine)
        event = df.to_dict(orient='records')[0]
        event['_metadata'] = {'input_label': input_label, 'table': table_name, 'event_type': table_name, 'input_type': 'odbc_psql_incremental_load'}

        # Turn timestamp into a string so testing in the API will work
        for k, v in event.items():
            if isinstance(v, datetime.datetime):
               event[k] = v.strftime('%Y-%m-%d %H:%M:%S.%f')

        if print_event:
            pprint.pprint(event, indent=2, width=200)

        return event



from setuptools import setup

setup(
    name='managealooma',
    version='0.1.14',
    author='Jay Rosenthal',
    author_email='jay.rosenthal@hover.to',
    packages=['managealooma'],
    url='http://pypi.python.org/pypi/ManageAlooma/',
    license='LICENSE.txt',
    description='Code to manage the Alooma ETL tool.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    zip_safe=False,
    install_requires=[
        "alooma>=0.4.15",
        "requests>=2.21.0",
        "tabulate>=0.8.3",
        "pandas>=0.23.4",
        "SQLAlchemy>=1.2.15"
    ],
    test_suite='nose.collector',
    tests_require=['nose'],
)



import pandas as pd
from tabulate import tabulate
import pprint


class System:

    def __init__(self, api):
        """ Gets information on system information like notifications, status codes, and metrics
        :param api: api authentication using the Alooma package
        """
        self.api = api

    def get_status_codes(self):
        """ Gets the possible status codes from Alooma

        :return: A dictionary with status codes and their descriptions
        """
        return self.api.get_samples_status_codes()

    def status_code_info(self, print_format='table', table_limit=None, pprint_indent=2, pprint_width=20, pprint_depth=5):
        """ Prints the status codes that Alooma may return from with the event info

        :param print_format: string 'table' to print event info as tables or 'json' to print as dictionaries
        :param table_limit: A limit to the columns to print
        :param pprint_indent: The indent value to pprint dictionaries
        :param pprint_width: The width value to pprint dictionaries
        :param pprint_depth: The depth value to pprint dictionaries
        :return: A dictionary with the status codes and their descriptions

        """

        status_codes = self.get_status_codes()

        if print_format == 'table':
            df = pd.Series(status_codes).reset_index()
            df.columns = ['Status Code', 'Status Description']
            print (tabulate(df.head(n=table_limit), headers='keys', tablefmt='psql', showindex=True))
        elif print_format == 'json':
            pprint.pprint(status_codes, indent=pprint_indent, width=pprint_width, depth=pprint_depth)

        return status_codes

    def system_metric_info(self, metric_type_names=None, last_n_minutes=60):
        """ Gets the system metrics and prints a dataframe with the results

        :param metric_type_names: string or list A list of systems metrics or a single metric name.
        :param last_n_minutes: The length of time in minutes counting back from the current time to retrieve notifications for
        :return: A dataframe with the inforamtion
        """

        metric_names = ['EVENTS_IN_PIPELINE',
                        'UNMAPPED_EVENTS',
                        'IGNORED_EVENTS',
                        'ERROR_EVENTS',
                        'LOADED_EVENTS_RATE']

        if not metric_type_names:
            metric_type_names = metric_names

        system_metrics = self.api.get_metrics_by_names(metric_names=metric_type_names, minutes=last_n_minutes)

        lst = []

        for metrics in system_metrics:

            row_dict = {'target': metrics['target'],
                        'value1':  metrics['datapoints'][0][0],
                        'timestamp1': pd.to_datetime(metrics['datapoints'][0][1], unit='s'),
                        'value2': metrics['datapoints'][1][0],
                        'timestamp2': pd.to_datetime(metrics['datapoints'][1][1], unit='s')}

            lst.append(row_dict)

        df = pd.DataFrame(lst)
        print(tabulate(df, headers='keys', tablefmt='psql', showindex=True))

        return df

